'''
This module contains the classes / functions used in the back-end of image processing (its front end is the second tab of the program).

Many classes / functions in this file are available in the public (non-GUI) API of PalmettoBUG.
'''
## License / derived-from info (commented out so as to not be lumped into the API docs)

#While the PalmettoBUG project as a whole is licensed under the GPL3 license, including this file, portions of this file ::
#
#        {  functions marked with >>>> # ****stein_derived (notes)  }
#                                                       
# are derivative / partially copy&paste with modification from the steinbock package (Copyright University of Zurich, 2021, MIT license)

# License text:
#---------------------------------------------------------------------------
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
#--------------------------------------------------------------------------------

# additionally the "Steinbock_Unhooked" module that is loaded into this script is ENTIRELY composed of modified steinbock package code
#
# additionally, throughout the design of this script, concepts / structures from CATALYST (https://github.com/HelenaLC/CATALYST/tree/main, GPL>=2 license) 
# and steinbock are used, such as the panel / metadata file structures, column names, usage, etc.

# see Assets / Other_License_Details.txt for more information on 3rd-party sources of code / ideas in this package.

import os
import warnings
from typing import Union
from pathlib import Path
import tkinter as tk

import customtkinter as ctk
import skimage
# from cellpose import denoise, models                                              
import tifffile as tf
import pandas as pd
import numpy as np

from .._vendor import readimc
from .._vendor.fcsy import DataFrame
from .._vendor import fcsparser
from .._vendor import pyometiff as pot

from .._vendor import steinbock as stein_unhook 
from ..Utils.sharedClasses import DirSetup, TableLaunch, Analysis_logger, Project_logger, warning_window  

'''
try:
    import navis as nv       ### currently, there is an error in navis on small meshes (I think?) so recommend to not use this  TODO: reactivate once fixed?
except Exception:
    pass
'''

__all__ = ["ImageAnalysis", 
            "mask_expand", 
            "imc_entrypoint", 
            "read_txt_file",
            "txt_folder_to_tiff_folder",
            "setup_for_FCS"]

_in_gui = False
def toggle_in_gui():
    global _in_gui
    _in_gui = not _in_gui
    return _in_gui

def _my_auto_hpf(image: np.ndarray[float], hpf: float = 0.75):
    '''
    Assumes the image's channels are in the first layer of the numpy array
    '''
    for ii,i in enumerate(image):
        hpf_threshold = np.quantile(i, hpf)
        #print(hpf_threshold)
        image[ii,:,:] = stein_unhook.filter_hot_pixels(i,hpf_threshold)
        
    return image

def imc_entrypoint(directory: Union[Path, str], 
                   resolutions: list[float] = [1.0,1.0], 
                   from_mcds: bool = True,
                   ) -> "ImageAnalysis":
    '''
    This function is the entrypoint for a project using MCDs or TIFFs. It initializes and return an ImageAnalysis object using the 
    arguments passed in by the user.

    Args:
        directory (Path or string): 
            This is the path to a folder containing a subfolder /raw with either .tiff or .mcd files inside it

        resolutions (iterable of length two: float, float): 
            This is the [X, Y] resolutions of the images in micrometers / pixel. 
            The default is 1.0 microns / pixels for both dimensions, as has been usual for IMC. 

        from_mcds (boolean): 
            whether the /raw subfolder contains .mcd files (= True) or .tiff files (= False)

    returns:
        a palmettobug.ImageAnalysis object
    '''
    directory = str(directory)
    try:
        resolutions[0] = float(resolutions[0])
        resolutions[1] = float(resolutions[1])
    except ValueError:
        print("Resolution X / Y must be numbers!")
        return
    directory = directory.replace("\\","/") 
    try:
        experiment = ImageAnalysis(directory, resolutions, from_mcds = from_mcds)
    except Exception as e:
        print(e)
        if from_mcds:
            if _in_gui:
                tk.messagebox.showwarning("Warning!", message = "Are you sure there are .mcd files in the 'raw' folder of your directory? \n" 
                            "Error in generating directory structure and preliminary panel file")
                return
            else:
                print("Are you sure there are .mcd files in the 'raw' folder of your directory? \n" 
                    "Error in generating directory structure and preliminary panel file")
                return 
             
        elif not from_mcds:
            if _in_gui:
                tk.messagebox.showwarning("Warning!", message = "Are you sure there are image files in the 'img' folder of your directory? \n" 
                        "Error in generating directory structure and preliminary panel file")
                return
            else:
                print("Are you sure there are image files in the 'img' folder of your directory? \n" 
                    "Error in generating directory structure and preliminary panel file")
                return      
    
    experiment.directory_object.makedirs()
    return experiment

def read_txt_file(path: Union[str, Path], num_meta_data_columns: int = 6):
    '''
    Reads a single txt file (these are backups for if mcd files become corrupted).

    Assumes that the txt file itself is not corrupted & has six [Arg: num_meta_data_columns] metadata columns before the channels.
    Also assumes a complete file with X / Y columns, whose values accurately correspond to the X / Y of the image to be generated.

    Args:
        path (str or Path): 
            the full path to the single .txt file to be read-in

         num_meta_data_columns (integer): 
            The number of metadata columns in the file. Default is 6, which worked for the files I tested this function on, 
            but I don't know what is standard / what kind of variability there is in the metadata for these types of files.

    returns:
        numpy array, which can be saved as an (.ome).tiff
    '''
    image_df = pd.read_csv(str(path), delimiter = "\t")
    image_channels_only = image_df.iloc[:,num_meta_data_columns:]
    image_array = np.array(image_channels_only).reshape([image_df['X'].max() + 1,
                                                         image_df['Y'].max() + 1, 
                                                         len(image_channels_only.columns)])
    image_array = image_array.transpose([2,0,1])
    return image_array

def txt_folder_to_tiff_folder(txt_folder: Union[Path, str], 
                              tiff_folder: Union[Path, str],
                              panel: pd.DataFrame, 
                              hpf: Union[float, int] = 50, 
                              ome_tiff_metadata: bool = False, 
                              resolutions: list[float] = [1.0,1.0], 
                              num_meta_data_columns: int = 6,
                              ) -> None:
    '''
    Iteratively runs read_txt_file() on each file in a directory, and output .ome.tiffs (at least by file extension) to an output directory.

    To actually generate ometiff metadata to be included with the images, two things must occur:

        1. ome_tiff_metadata == True

        2. Resolutions Argument must be provided (default is 1 micrometer/pixel in both X an Y)

    Args:
        txt_folder (Path or str): 
            the path to a folder where the .txt files to be converted are. Assumes only .txt files are inside this folder

        tiff_folder (Path or str): 
            the path to the folder where the output .ome.tiff files will be written

        panel (pandas DataFrame): 
            the panel file. Used to filter out undesirable channels & construct ometiff metadata

        hpf (integer >= 0): 
            whether to run hot pixel filtering (if hpf != 0) and if hpf is run what threshold to use. 

        ome_tiff_metadata (boolean): 
            whether to generate metadata for the ome.tiff -- this does not change the output extension name, which will always be .ome.tiff, but 
            does determine whether the function will attempt to add metadata to the files. If True, then the resolutions argument must be provided, or an error will occur. 

        resolutions (list of two floats > 0): 
            the resolution of the images in X and Y dimensions, in micrometers per pixel. Only used if ome_tiff_metadata == True. 

        num_meta_data_columns (integer):
            The number of metadata columns in the file. Default is 6, which worked for the files I tested this function on, 
            but I don't know what is standard / what kind of variability there is in the metadata for these types of files. 

    Inputs / Outputs:
        Inputs: 
            reads each file inside txt_folder (expecting all to be .txt files of the format exported during IMC as backups of .mcd files. )

        Outputs: 
            writes an .ome.tiff file in tiff_folder for each .txt file read in from txt_folder
    '''
    txt_folder = str(txt_folder)
    tiff_folder = str(tiff_folder)
    txt_files = ["".join([txt_folder, "/", i]) for i in sorted(os.listdir(txt_folder)) if i.lower().find(".txt") != -1]
    tiff_files_out = ["".join([tiff_folder, "/", i.rstrip("txt"),"ome.tiff"]) for i in sorted(os.listdir(txt_folder)) if i.lower().find(".txt") != -1]
    img_slicer = (panel['keep'] == 1)
    for i,ii in zip(txt_files, tiff_files_out):
        image = read_txt_file(i, num_meta_data_columns = num_meta_data_columns)[img_slicer,:,:]
        if (hpf > 0) and (hpf < 1):
            image = _my_auto_hpf(image, hpf)
        elif hpf != 0:
            image = stein_unhook.filter_hot_pixels(image,hpf)
        if ome_tiff_metadata:
            ome_metadata = _generate_ome_tiff_metadata(output_directory = ii[:-9],
                                                       filename = ii[ii.rfind("/") + 1:], 
                                                       resolutions = resolutions, 
                                                       panel_csv = panel,
                                                       )
            write_ome_tiff(image, ome_metadata, ii)
        else:
            tf.imwrite(ii, image)

def launch_denoise_seg_program(directory: Union[Path, str], 
                               resolutions: list[float] = [1.0, 1.0],
                               ) -> None:
    '''
    This launches the 'isoSegDenoise' sister program, with the provided directory pre-loaded, presuming that the isoSegDenoise package is
    installed in the same python environment as PalmettoBUG.

    Args:
        directory (str or Path): 
            the directory to launch isoSegDenoise inside of

        resolutions (list of two floats > 0): 
            the resolution of the images in X and Y dimensions, in micrometers per pixel.
    '''
    directory = str(directory)
    if not os.path.exists(directory):
        return
    res1 = float(resolutions[0])
    res2 = float(resolutions[1])
    import subprocess
    try:
        subprocess.run(['segdenoise','-d', f'{directory}', '-r1', f'{str(res1)}', '-r2', f'{str(res2)}'])
        # Useful discussion for switching from shell = True to shell = False & using list[str] instead of string (as I had originally done)
        #       https://stackoverflow.com/questions/3172470/actual-meaning-of-shell-true-in-subprocess
    except Exception:
        print("Error in launching the Segmentation / Denoising process")

def _MCD_Generator(MCD_list: list[Path]):   # ****stein_derived 
                                            # (a substantially sipmlified / stripped down version of 
                                            #       steinbock.preprocessing.imc._try_preprocess_mcd_images_from_disk)
    '''
    A much simplified MCD reader compared to steinbock's, simiarly based on readimc.
    This is a helper function for when converting /raw folder MCDs to .ome.tiffs in the /img folder
    '''
    for i in MCD_list:
        with readimc.MCDFile(i) as mcd:
            for j in mcd.slides:
                for k in j.acquisitions:
                    try:
                        image = mcd.read_acquisition(k)
                        path = i
                        ROI = k.description
                        yield image, path, ROI, k
                    except Exception:
                        print(f"A ROI in the following MCD: \n {str(i)} \n failed to read!")

def _TIFF_Generator(TIFF_list: list[Path]):   # ****stein_derived 
                                                # (ish -- only in the sense of its structure: 
                                                # similar input, output, and generator design as _MCD_Generator)
    '''
    Similar to _MCD_Generator() above, this function is a helper for when converting from /raw folder Tiffs --> .ome.tiffs in the /img folder
    '''
    for i in TIFF_list:
        i = str(i)
        reader =  pot.OMETIFFReader(i)
        img_array, metadata, xml_metadata = reader.read()
        path = i[:i.rfind("/")]
        ROI = i[i.rfind("/"):]
        yield img_array, path, ROI, metadata

def _generate_ome_tiff_metadata(panel_csv: pd.DataFrame, 
                                output_directory: str, 
                                filename: str, 
                                resolutions: list[float],
                                ) -> dict:
    '''
    Generates .ome.tiff metadata for writing to .ome.tiff file type. (Helper function when converting from MCDs --> .ome.tiffs)
    
    Requires the panel_csv, an image, the output directory (like directory + '/img') and the file name
    The file name should not have the entire directory nor the suffix (as in, "C:/whatever/path/" and ".mcd" should NOT be included).
    The image should have the following order of dimensions: channels, Y, X (at least for current implementation, 
    update to be accurate after testing).

    resolutions are length per pixel in X and Y direstions (in uM)

    The fields chosen for the metadata were taken from a .ome.tiff file with multi-channels, 
    from a publicly available IMC experiment on Zenodo (should be: https://zenodo.org/records/8023452,
    Or perhaps instead was from the tumor-sphere dataset from this paper by the Bodenmiller group: 
                                                                    Zanotelli, Vito Rt et al. 
                                                                    “A quantitative analysis of the interplay of environment, neighborhood, 
                                                                    and cell state in 3D spheroids.” 
                                                                    Molecular systems biology vol. 16,12 (2020): e9798. 
                                                                    doi:10.15252/msb.20209798). 
    '''
    ome_metadata = {}
    ome_metadata['Directory'] = str(output_directory)
    ome_metadata['Filename'] = filename
    ome_metadata['Extension'] = 'ome.tiff'
    ome_metadata['ImageType'] = 'ometiff'
    ome_metadata["PhysicalSizeX"] = resolutions[0]
    ome_metadata["PhysicalSizeXUnit"] = "micrometer"
    ome_metadata["PhysicalSizeY"] = resolutions[1]
    ome_metadata["PhysicalSizeYUnit"] = "micrometer"
    panel_csv = panel_csv[panel_csv['keep'] == 1].reset_index()
    channel_dict = {}
    for i,ii in enumerate(panel_csv['name']):
        channel_dict[ii] = {'Name':ii,'ID':str(i),'Fluor':str(panel_csv['channel'][i])}
    ome_metadata['Channels'] = channel_dict
    return ome_metadata

## originally made to test concurrent futures multi-threading of the I/O tasks. Turns out, multi-threading is quite inefficient....
def write_ome_tiff(image, ome_tiff_metadata, file_path) -> None:
    '''
    Takes in a image as a numpy array, metadata for the image (in the same format that pyometiff generates when reading an ome-tiff), and a filepath
    where the image will be written as an .ome.tiff 
    '''
    #image = image[:,np.newaxis,np.newaxis,:,:]
    writer = pot.OMETIFFWriter(
            fpath=file_path,
            dimension_order='CYX',   # 'CZTYX'
            array=image,
            metadata=ome_tiff_metadata,
            explicit_tiffdata=False)
    writer.write()

def mask_expand(distance: int, 
                image_source: Union[Path, str], 
                output_directory: Union[Path, str, None],
                ) -> None:                                   # ****stein_derived (ish -- 
                                                                # only in the sense that it conciously replicates a particular steinbock utility 
                                                                # using the same skimage function.  The actual implementation here is not very 
                                                                # similar to steinbocks')
    '''
    Function that expands the size of cell masks:

    Args:
        distance (integer):  
            the number of pixels to expand the masks by

        image_source (string or Path): 
            the file path to a folder containing the cell masks (as tiff files) to expand

        output_directory (string or Path): 
            the file path to a folder containing the cell masks (as tiff files) to expand

    Inputs / Outputs:
        Inputs: 
            reads every file in image_source folder (expecting .tiff format for all)
    
        Outputs: 
            writes a (.ome).tiff into the output_directory folder for each file read-in from image_source
            The filenames in the image_source folder as preserved in the output_directory, so if image_source == output_directory
            then the original masks will be overwritten. 
    '''
    from skimage.segmentation import expand_labels
    output_directory = str(output_directory).rstrip("/")
    image_source = str(image_source)
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)

    for i in sorted(os.listdir(image_source)):
        if i.lower().find('.tif') != -1:
            read_dir = "".join([image_source, "/", i])
            out_dir = "".join([output_directory, "/", i])
            read_in_mask = tf.imread(read_dir)
            expanded_mask = expand_labels(read_in_mask, distance)
            tf.imwrite(out_dir, expanded_mask, photometric = "minisblack")

class ImageAnalysis:
    '''
    This handles Image Processing steps of PalmettoBUG, such as conversion from mcd's and segmentation measurements

    Args:
        directory (str, Path, or None):
            The directory to step up the image analysis / PalmettoBUg project inside of. Is expecting the directory to already exist & there
            to be .mcd or .tif / .tiff files in a /raw subfolder of directory.
            If None, then initiates an ImageAnalysis object without needing to set up the directory. This can be useful if you don't need the 
            /raw --> /images/img conversion step and you intend to manually set the input / output folder paths at each step.

        resolutions (list of float): 
            Default = [1.0, 1.0]. Represents the width of the pixels in [X, Y] directions in micrometers.

        from_mcds (bool): 
            If True, assumes that there will be MCD files in /raw. If False, presumes there are .tif/.tiff files in /raw. 

    Key Attributes:
        directory (str): 
            the path to a folder containing a /raw/ subfolder where the MCD or TIFF files are

        directory_object (DirSetup): 
            this attribute is a sub-class from Utils/SharedClasses.py module. it coordinates directories of the typical
            PalmettoBUG project --

            *Key Subattributes*

            - (self.directory_object).main == self.directory >>> The highest-level directory of the project

            - raw_dir == {directory}/raw >>> the folder containing the MCD or TIFF files of raw data

            - img_dir == {directory}/images >>> the folder containing sub-folders of images (such as the img_dir/img sub-folde, which contains
                the initial images directly converted from the raw_dir)

            - masks_dir == {directory}/masks >>> the folder containing sub-folders of cell masks (such as the masks_dir/deepcell_masks sub-folder, 
                which contains the cell masks created by deepcell, before any modifications / expansions are performed on those masks.)

            - classy_masks_dir == {directory}/classy_masks >>> the folder containing sub-folders of classy cell masks. Each subfolder is named
                by convention using the pixel classifier + cell mask pair that the clasy masks were derived from.

            - px_classifiers_dir == {directory}/Pixel_Classification >>> the folder containing sub-folders of pixel classifiers

            - Analyses_dir == {directory}/Analyses >>> the folder containing sub-folders of Analysis directories. 
            
            - logs == {directory}/Logs --> the folder containing the .log files generated by the GUI

        from_mcds (boolean): 
            whether the files in directory/raw are MCD files (True), or are TIFF files (False)

        resolutions (list[float, float]): 
            the X and Y resolutions of the images, in micrometers / pixel

        panel (pandas dataframe): 
                This is a pandas dataframe read-in from & written to directory/panel.csv
                This is a steinbock-style panel (with changes), with four columns = "name","antigen","keep","segmentation"

        metadata (pandas dataframe): 
                This is the dataframe containing a CATALYST-style metadata file. It is really intended for the 
                Analysis portion of the program and is not used in image processing, however this class produces a preliminary
                version of this dataframe as a part of the transition from image processing --> analysis

                It has four columns -- 

                    "file_name" == the filenames of the .fcs files in the analysis

                    "sample_id" == numbers to identify each filename quickly (zero-indexed)

                    "patient_id" == a secondary grouping / covariate / batch

                    "condition" == the independent variable / treatment vs. control grouping

        Analysis_panel (pandas dataframe): 
                This is the dataframe containing a CATALYST-style panel file. It is really intended for the 
                Analysis portion of the program and is not used in image processing, however this class produces a preliminary
                version of this dataframe as a part of the transition from image processing --> analysis

                It has three columns:

                    "fcs_colname" == the name of the marker in the .fcs files. when coming straight from solution-mode fcs files, 
                    these names can be non-straightforward or confusing (often metal names. Ex: Ce140Di) When produced from an 
                    imaging experiment the fcs names are usually identical to the "antigen" names:

                    "antigen" == the name of the marker to use in analysis plots, usually its straightforward, biological name (ex: CD4)

                    "marker_class" == 'type','state', or 'none' -- used as in CATALYST-style workflow to determine how markers are used.

    '''
    def __init__(self, directory: Union[Path, str, None], resolutions: list[float,float] = [1.0, 1.0], from_mcds: bool = True):
        '''
        Directory can be set to None, in order to not set up the panel / directory. This allows this class to initialized without data, which can be 
        useful if not intending following the standard PalmettoBUG directory structure. X and Y are the resolution of the images (in micrometers), 
        and from_mcds indicates whether the /raw files are mcds (True) or .tiffs (False)
        '''
        self.from_mcds = from_mcds
        self.resolutions = resolutions   ## load these values later   (default of 1uM is assumed for now)
        self.metadata = pd.DataFrame()
        self.Analysis_panel = pd.DataFrame()
        if directory is not None:
            directory = str(directory)
            self.directory = directory
            self.directory_object = DirSetup(directory)
            self._panel_setup()

    def _panel_setup(self) -> None:
        '''
        This will either read in the panel.csv file in the top-level project folder, or failing that will attempt to generate a panel file automatically.

        Note:
            -- This method depends on the self.from_mcds attribute to know whether mcds or tiffs are in the /raw folder

            -- Any automatically generated panel file will have an entirely blank 'segmentation' column. This requires editing 
            before segmentation can be performed using deepcell or cellpose

            -- additionally, the automatically generated 'keep' column may not accurately reflect what you want, although it often does. 
        '''
        if self.from_mcds is True:
            MCD_list = stein_unhook.list_mcd_files(self.directory_object.main)
            try:
                self.panel = pd.read_csv("".join([self.directory_object.main, "/panel.csv"]))
            except FileNotFoundError:
                self.panel = stein_unhook.create_panel_from_mcd_files(MCD_list)
                self.panel = self.panel.drop(['ilastik','cellpose','deepcell'], axis = 1)     
                                    ######## remove unwanted columns from the underlying steinbock package implementation
                self.panel['segmentation'] = ""

                ## This auto-sets background channels to keep = 0 (based on duplicating entry in the channel / name columns)
                numbers_channel = self.panel['channel'].str.replace("[^0-9]","", regex = True)
                letters_channel = self.panel['channel'].str.replace("[0-9]","", regex = True)
                self.panel['channel_test'] = numbers_channel + letters_channel

                numbers_name = self.panel['name'].str.replace("[^0-9]","", regex = True)
                letters_name = self.panel['name'].str.replace("[0-9]","", regex = True)
                self.panel['name_test'] = numbers_name + letters_name
                
                keep = (self.panel['channel_test'] != self.panel['name_test'])
                self.panel['keep'] = keep
                self.panel['keep'] = self.panel['keep'].astype('int')
                self.panel = self.panel.drop(['channel_test','name_test'],axis=1)
        else:  ## if self.mcds is False, then loading from .tiffs
            try:  ## read in panel file if it already exists
                read_dir = "".join([self.directory_object.main, "/panel.csv"])
                self.panel = pd.read_csv(read_dir)      
            except FileNotFoundError:
                image_list = sorted(os.listdir(self.directory_object.main + "/raw"))
                image_list = [i for i in image_list if i.lower().find(".tif") != -1]   ## exclude any non- .tif(f) files  (looking at you, .DS_store ......)
                reader_string = "".join([self.directory_object.main, "/raw/", image_list[0]])
                if image_list[0][-9:].lower() == ".ome.tiff":   
                    # when dealing with an .ome.tiff, I'd like to try to recover some useful metadata (only uses first image):
                    reader = pot.OMETIFFReader(reader_string) 
                    img_array, metadata, xml_metadata = reader.read()
                    try: 
                        channel_list = [i for i in metadata['Channels']]
                    except KeyError:
                        tiff_file1 = tf.imread(reader_string)
                        channel_list = [i for i in range(tiff_file1.shape[0])]
                else:            # this means we are dealing with a .tiff, and will not try to recover metadata
                    tiff_file1 = tf.imread(reader_string)
                    channel_list = [i for i in range(tiff_file1.shape[0])]
                #### now make the initial pd.DataFrame to hold the table widget:
                Init_Table = pd.DataFrame()
                Init_Table['channel'] = channel_list
                Init_Table['name'] = channel_list
                Init_Table['keep'] = 1         # default to keeping all channels
                Init_Table['segmentation'] = "" 
                self.panel = Init_Table
                Init_Table.to_csv(self.directory_object.main + "/panel.csv", index = False)   

    def panel_write(self) -> None:
        '''
        This method writes the self.panel dataframe to the disk at the expected location for future re-read 
        (panel.csv in the top-level folder of the project) 
        '''
        if _in_gui:
            try:
                self.panel.to_csv(self.directory_object.main + '/panel.csv', index = False)
                with open(self.directory_object.main + '/panel.csv') as file:
                    Project_logger(self.directory_object.main).return_log().info(f"Wrote panel file, with values: \n {file.read()}")
            except Exception:
                tk.messagebox.showwarning("Warning!", message = """Could not write panel file! \n 
                            Do you have the .csv open right now in excel or another program?""")
        else:
            self.panel.to_csv(self.directory_object.main + '/panel.csv', index = False)
            
     # This method writes from MCDs --> .ome.tiffs 
    def raw_to_img(self, 
                   hpf: int = 50, 
                   input_directory: Union[str, Path, None] = None, 
                   output_directory: Union[str, Path, None] = None,
                   ) -> None:
        '''
        This method converts/moves files from the /raw folder --> /images/img folder. It always exports in .ome.tiff format
        with two transformations of the images in the raw files:
        
            1. channels with 'keep' == 0 in the panel file will be dropped from the exported .ome.tiffs

            2. hot pixel filtering will be performed before exporting to the /images/img folder if hpf > 0

        It depends on self.from_mcds to know whether to expect MCD or TIFF files in the /raw folder.

        Args:
            hpf (int >= 0): 
                an integer denoting the threshold used for steinbock-style hot pixel filtering. This means that pixels
                that are brighter than each of their surrounding neighbor pixels by more than the inputted threshold 
                will have their values reduced to match the value of their brightest neighbor pixel. So lower 
                thresholds will more aggressively filter "hot" / bright pixels, while higher thresholds will filter less. 
                The default (50) matches the default value in steinbock. ** If hpf == 0, no hot pixel filtering will occur. **

            input_directory (str, Path, or None): 
                a path to the folder containing the .mcd or .tiff files to be converted and 
                hot-pixel filtered. Assumes that the folder chosen ONLY contains files (no sub-folders) and 
                that all of those files are of the appropriate format.

            output_directory (str, Path, or None): 
                A path to a folder to write the .ome.tiff files to. If None, then defaults 
                to the self.directory.main + "/images/img/" subfolder

        Returns:
            None: (its output is in writing to the disk, not returning a value)
        '''
        from_mcds = self.from_mcds
        if input_directory is None:
            input_directory = self.directory_object.main + "/raw/"
        else:
            input_directory = str(input_directory)

        if output_directory is None:
            output_directory =  "".join([self.directory_object.img_dir, '/img/'])
        else:
            output_directory = str(output_directory)
        if not os.path.exists(output_directory):
            os.mkdir(output_directory)

        if from_mcds is True:
            mcd_list = ["".join([input_directory,i]) for i in sorted(os.listdir(input_directory)) if i.lower().find(".mcd") != -1]
            MCD_gen = _MCD_Generator(mcd_list)
        else:
            tiff_list = ["".join([input_directory,i]) for i in sorted(os.listdir(input_directory)) if i.lower().find(".tif") != -1]
            MCD_gen = _TIFF_Generator(tiff_list)

        i = 1
        counter = -1
        keep = (self.panel['keep'] == 1)
        while i == 1:
            try: 
                image, path, ROI, acquisition = next(MCD_gen)
                counter += 1
            except StopIteration:
                break
            # filter for keep channels
            if len(image) != len(self.panel):
                if _in_gui:
                    warning_window(f"""The number of channels in {ROI} of {path} does not match the number of channels 
                                   in the panel file! Skipping this ROI.""")
                else:
                    print(f"""The number of channels in {ROI} of {path} does not match the number of channels 
                          in the panel file! Skipping this ROI.""")
            else:   # proceed with conversion to TIFF
                image = image[keep]
                if (hpf > 0) and (hpf < 1):
                    image = _my_auto_hpf(image, hpf)
                elif hpf != 0:
                    image = stein_unhook.filter_hot_pixels(image, hpf)
                if from_mcds is True:
                    file_name = "".join([(str(path)[len(self.directory_object.main)+5:-4]), '_', ROI])   ### this is an awkard line -- 
                                                                                                        # len(directory) + 5, slices off the 
                                                                                                        # directory + /raw/ ,
                                                                                                        #  while -4 slices off the .mcd suffix
                    out = "".join([output_directory, file_name, ".ome.tiff"])
                    ome_tiff_metadata = _generate_ome_tiff_metadata(self.panel, 
                                                                    self.directory_object.img_dir, 
                                                                    file_name, 
                                                                    self.resolutions)
                    write_ome_tiff(image, ome_tiff_metadata, out)
                else:
                    if ROI.rfind(".tiff") == -1:   ## this means that the current format must be .tif (one f, instead of two), so I add an "f" for consistency
                        ROI = ROI + "f"
                    out = "".join([output_directory, ROI])
                    ome_tiff_metadata = acquisition
                    if ome_tiff_metadata is None:
                        file_name = ROI[1:]
                        ome_tiff_metadata = _generate_ome_tiff_metadata(self.panel, 
                                                                    self.directory_object.img_dir, 
                                                                    file_name, 
                                                                    self.resolutions)
                    write_ome_tiff(image, ome_tiff_metadata, out)

    def instanseg_segmentation(self, 
                               re_do: bool = False,
                               input_img_folder: Union[Path, str, None] = None, 
                               single_image: Union[Path, str, None] = None,
                               output_mask_folder: Union[Path, str, None] = None,
                               channel_slice: Union[None, np.array] = None,
                               merge_channels: bool = False,
                               pixel_size: Union[float, None] = None,
                               target: str = "cells",
                               mean_threshold: float = 0.0,
                               model: str = "fluorescence_nuclei_and_cells") -> None:
        '''
        Instanseg is an open-source (no non-commercial issues) deep-learning segmentation algorithm: https://github.com/instanseg/instanseg/tree/main

        Channels are scaled by a min_max transformation before being used (and before being merged together, if that is chosen).

        Args:
            input_img_folder (str, Path, or None):
                The path to a folder contianing the images to segment. If None, defaults to f"{self.directory_object.img_dir}/img"
            
            output_mask_folder (str, Path, or None):
                The path to a folder where you want the segmentation masks to be written to. If None, defaults to f"{self.directory_object.masks_dir}/instanseg_masks"

            single_image (str, Path, or None):
                If not None (and not the empty string ""), this parameter provides the name or path of a single image in the input_img_folder to segment.

            channel_slice (integer numpy array or None):
                If provided, will be used to slice each image array to subset the channels provided to the instanseg model. The length of this array
                must be the same as the number of channels in the images.
                Specifically, the channels in the image that will be used will be:  image[channel_slice > 0]
                If None, all channels in all images are used as independent channels.

            merge_channels (boolean):
                IF channel_slice is provided, this determines whether the selected channels are merged into two (cytoplasmic / nuclear -- True) or left
                as separate channels (False, default).

            pixel_size (float or None):
                resolution of the pixels in the images. If None, defaults to using self.resolutions (self.resolutions[0] == self.resolutions[1] must be true)
                Provided to the pixel_size argument of the instanseg model 

            target (str):
                "cells", "nuclei", or "all_outputs". Whether to try to segment whole cells, only the nuclei, or both. 
                Provided to the target argument of the instanseg model 

            mean_threshold (float):
                Higher values decrease the number of cells (higher threshold for identifying a cell) while lower number should increase the number of 
                detected cells

            model (str):
                what pre-trained instanseg model to use. Currently instanseg only offeres two models (fluoresence-based, the default, and a H&E based model)
                More options will hopefully open up as this segmentation model is developed. Theoretically, there should be a way to allow custom-trained
                models to loaded as well, which could be quite nice.

        ## example test script  -- results so far: it works in that it runs, but the results are very poor compared to deepcell
        ## maybe instanseg needs a dedicated IMC model, or needs a larger training set (like TissueNet, but that particular dataset would create 
        ## license issues))
        import palmettobug
        proj_dir = f"{my_computer_path}/Example_IMC"
        image_object = palmettobug.ImageAnalysis(proj_dir, resolutions = [1.0,1.0])
        panel_keep_only = image_object.panel[image_object.panel['keep'] == 1]
        nuclei_slice = panel_keep_only['segmentation'] == "Nuclei"
        image_object.instanseg_segmentation(channel_slice = nuclei_slice, target = "nuclei", mean_threshold = -1.0)
        '''
        from instanseg import InstanSeg
        if input_img_folder is None:
            input_img_folder  = f"{self.directory_object.img_dir}/img"

        if output_mask_folder is None:
            output_mask_folder  = f"{self.directory_object.masks_dir}/instanseg_masks"
            if not os.path.exists(output_mask_folder):
                os.mkdir(output_mask_folder)
        
        if pixel_size is None:
            if self.resolutions[0] != self.resolutions[1]:
                print("Error! Pixel resolutions are inconsistent between X and Y! Cancelling Instanseg segmentation")
                return
            pixel_size = self.resolutions[0]

        model = InstanSeg(model)

        source_images = [i for i in os.listdir(input_img_folder) if i.lower().find(".tif") != -1]
        if (single_image is not None) and (single_image != ""):
            single_image = str(single_image)
            path_or_name = single_image.find("/")
            if path_or_name != -1:   ## if single_image is a path, convert it to just the filename
                single_image = single_image[(path_or_name + 1):]
            if single_image not in source_images:
                raise ValueError(f"No valid .tif filename -- {single_image} -- was found in the supplied input folder: {input_img_folder}")
            source_images = [single_image]
            
        if not re_do:
            existing = os.listdir(output_mask_folder)
            source_images = [i for i in source_images if i not in existing]

        for i in source_images:
            image_array = tf.imread(f'{input_img_folder}/{i}')
            for j,jj in enumerate(image_array):
                image_array[j] = (jj - jj.min()) / (jj.max() - jj.min())  ## min-max scale all channels -- channels are assumed to be the first dimension of the array!
            if channel_slice is not None:
                kept_channel_slice = channel_slice[channel_slice > 0]
                if len(kept_channel_slice) == 0:
                    print("A channel_slice was provided, but no channels were selected in it! Cancelling instanseg segmentation")
                    return
                if merge_channels:
                    unique_kept_channels = np.unique(kept_channel_slice)
                    if len(unique_kept_channels) != len(kept_channel_slice):
                        new_image_array = np.zeros([len(unique_kept_channels), image_array.shape[1], image_array.shape[2]])
                        for k,kk in enumerate(unique_kept_channels):
                            slicer = (channel_slice == kk)
                            new_image_array[k,:,:] = np.sum(image_array[slicer], axis = 0) / slicer.sum()
                        image_array = new_image_array.copy()
                else:
                    image_array = image_array[channel_slice > 0,:,:]
                image_array = image_array[(channel_slice > 0), :, :]            
            prediction = model.eval_medium_image(image_array, mean_threshold = mean_threshold, target = target, pixel_size = pixel_size)
            tf.imwrite(f'{output_mask_folder}/{i}', np.squeeze(np.asarray(prediction[0])))

    def mask_intersection_difference(self, 
                                    masks_folder1: Union[str, Path], 
                                    masks_folder2: Union[str, Path], 
                                    kind: str = 'intersection1', 
                                    object_threshold: int = 1, 
                                    pixel_threshold: int = 1, 
                                    re_order: bool = True, 
                                    output_folder: Union[None, str, Path] = None):
        '''
        Provide two folders of masks, and derive a third folder of masks from them transformed in some way. Masks are dropped as a whole (not pixel-wise),
        and there are a limited set of possible transformations:

             intersection1 (one-way) -- This keeps the masks from the first folder of masks, but only the masks that overlap with sufficient masks from folder2
                                        No masks from folder2 carry over to the output folder

             intersection2 (two-way) -- This keeps masks from both folders, as long as they overlap sufficiently. HOWEVER, masks from folder1 take precedence
                                        As in, where overlap exists only mask1 values will be carried over first and mask2 values will only end up in the output
                                        after that where the output has values of 0 (which is to say, only in regions outside the remaining masks from mask1)
                                        Additionally, masks from the second folder are given a value = mask2 + max(mask1) in the output so that they will remain distinct
                                        from folder1-derived masks. 

             difference1 (one-way) -- This keeps masks from folder1 only if a sufficient number of masks from folder2 do NOT overlap with them

             difference2 (two-way) -- This keeps masks from both folders, but only if they do not overlap with sufficient masks from the opposite folder. 
                                    HOWEVER: Masks from the first folder take precedence over mask from the second! 
                                    As in, the masks from folder1 which are kept in the transformation are carried over into the output first, and after that the 
                                    masks from folder2, but only into pixels with value 0 in the output. 
                                    This precedence should only matter if the thresholds are increased above the defaults of 1, as otherwise there should be no
                                    overlap at all between the saved masks from the two folders.
                                    Additionally, masks from the second folder are given a value = mask2 + max(mask1) in the output so that they will remain distinct
                                    from folder1-derived masks. 


        an overlapping mask is determined by the pixel_threshold value -- a mask from folder 2 is considered to overlap with a mask from folder 1 if the number of overlapping
            pixels between the two is greater than or equal to the pixel_threshold value. 

        'sufficient masks' is determined by the object threshold (default = 1, as in, just 1 overlapping mask form folder2 within a mask from folder 1 means 
            triggers the transformation) 

        Together, this should allow this function to be used to do things like only keeping cell masks within a particular region of the tissue or only keeping tissue 
        regions with sufficient number of cell masks inside them, etc. Or, by chaining this operation together, only keeping cells within particular region of tissue, where 
        those regions of tissue have sufficient numbers of cells (of a particular cell type, even, if using classy masks to further sophisticate things).
        The (possible) addition of this function was inspired by analyses performed in the following paper using pancreatic islets: 
            Damond, Nicolas et al. “A Map of Human Type 1 Diabetes Progression by Imaging Mass Cytometry.” Cell metabolism vol. 29,3 (2019): 755-768.e5. 
            doi:10.1016/j.cmet.2018.11.014
        The publicly-available data from this paper is also planned to be analyzed in PalmettoBUG, in order to compare the effectiveness of PalmettoBUG at 
        replicating prior work.
        
        Args:
            masks_folder1 / 2 (string, Path): 
                paths to two folders of masks -- as in, each folder is expected to contain single-channel, integer-valued tiff files where each integer represents
                a unique cell (or other object). There must be files in each folder with matching file names -- only these can be processed!
                Note that the order of the folders (as in, which is masks_folder1 vs masks_folder2) is very important for some transformations!

            kind (string):
                One of ['intersection1', 'intersection2', 'difference1', or 'difference2']. Determines how the maasks are transformed. See description above for details.
                Note that when kind = 'difference2', the object/pixel threshold comparisons are also utilized in the reverse (from mask folder 2 --> 1)

            object_threshold (integer):
                when determining whether a mask from folder1 overlaps with masks from folder2, this determines how many 'overlapping' objects inside it are sufficient
                to trigger keeping / discarding the mask. The default is 1, meaning that even a single overlapping mask is sufficient to trigger the transformation. 
                

            pixel_threshold (integer): 
                when determing whether a mask from folder2 overlaps with a mask from folder1, this determines how many pixels of mask2 is sufficient to consider
                it overlapping. When == 1 (default), this means that even a single pixel of overlap will count mask2 as an object inside mask1. The total count
                of such overlapping mask2 objects inside mask1 are then compared to the object_threshold to determine whether to keep / discard mask1 from the output.

            re_order (boolean):
                Whether to re-index the masks, starting from 1 and continuously increasing in increments of 1, so that there are no gaps / discontinuities in the values.
                Default = True, which re-indexes to start from 1 etc. However, if you want to preserve the original mask values (of mask1 only), so that they can be 
                matched to the original masks, set this parameter == False. Because of how two-way methods work, the original values of mask folder2 are not preserved
                regardless of this parameter.

            output_folder (string, Path, or None):
                The path to a file folder where the output, transformed masks can be written. If None (default), then the file folder name is automatically derived
                from the names of folder1 and folder2 (specifically: {self.directory_object.masks_dir}/{folder1}_{folder2} ). This folder automatically inside the masks
                directory of the PalmettoBUG project folder. 
                If provided, should be a FULL path to a create-able folder where the masks will be written. 

        Returns:
            None     (does, however, read & write .tiff files)
        '''
        masks_folder1 = str(masks_folder1)
        masks_folder2 = str(masks_folder2)
        masks1 = os.listdir(masks_folder1)
        masks2 = os.listdir(masks_folder2)
        matching_files = [i for i in masks1 if (i in masks2) and (i.rfind(".tif") != -1)]  ## only want .tif(f) files present in both folders
        if len(matching_files) == 0:
            print("Error: No filenames shared between the two folders of masks! Cancelling")
            return

        if output_folder is None:
            first_half = masks_folder1[masks_folder1.rfind("/") + 1:]
            second_half = masks_folder2[masks_folder2.rfind("/") + 1:]
            output_folder = self.directory_object.masks_dir + f"/{first_half}_{second_half}"
        if not os.path.exists(output_folder):
            os.mkdir(output_folder)

        for i in matching_files:
            mask1 = tf.imread(f'{masks_folder1}/{i}').astype('int32')
            mask2 = tf.imread(f'{masks_folder2}/{i}').astype('int32')
            if mask1.shape != mask2.shape:
                print(f"Warning! Mask file: {i} did  not have a matching shape between the two folders of masks. Skipping this file!")
            else:
                output = self._mask_bool(mask1, mask2, kind = kind, object_threshold = object_threshold, pixel_threshold = pixel_threshold)
                tf.imwrite(f'{output_folder}/{i}', output.astype('int32'))
        
    def _mask_bool(self, mask1: np.ndarray[int], 
                   mask2: np.ndarray[int], 
                   kind: str = 'intersection1', 
                   object_threshold: int = 1, 
                   pixel_threshold: int = 1, 
                   re_order: bool = True) -> np.ndarray[int]:
        ''' 
        helper for self.boolean_mask_transform, executing the operation on a single pair of masks
        '''
        if (kind =="difference2") or (kind =="intersection2"):
            backup = mask1.copy()
        mask_values = [i for i in np.unique(mask1) if i > 0]
        for j in mask_values:
            temp = mask2[mask1 == j]      ## look at mask2 with each mask of mask1, and count overlapping values
            overlapping_values = [i for i in np.unique(temp) if i > 0]
            object_counter = 0
            for k in overlapping_values:
                if (temp == k).sum() > pixel_threshold:
                    object_counter += 1

            if kind == "intersection1":
                if object_counter < object_threshold:
                    mask1[mask1 == j] = 0   ## delete masks from mask1 that do not have sufficient overlap with objects from mask2

            if kind == 'difference1':
                if object_counter > object_threshold:
                    mask1[mask1 == j] = 0   ## delete masks from mask1 that have sufficient overlap with objects from mask2

            if kind == "difference2":
                if object_counter > object_threshold:
                    mask1[mask1 == j] = 0
                
        if (kind =="difference2") or (kind =="intersection2"):
            mask_values = [i for i in np.unique(mask2) if i > 0]     ## if two-way difference, repeat the process but look from mask2 --> mask1 instead, then add kept mask2 to output
            for j in mask_values:
                temp = backup[mask2 == j]      
                overlapping_values = [i for i in np.unique(temp) if i > 0]
                object_counter = 0
                for k in overlapping_values:
                    if (temp == k).sum() > pixel_threshold:
                        object_counter += 1
                if kind == "difference2":
                    if object_counter < object_threshold:
                        mask1[(mask2 == j)*(mask1 == 0)] = j + np.max(backup)   ## add mask from mask2 --> mask1 (which is also the output), but only into 0-value pixels
                if kind == "intersection2":
                    if object_counter > object_threshold:
                        mask1[(mask2 == j)*(mask1 == 0)] = j + np.max(backup)   ## add mask from mask2 --> mask1 (which is also the output), but only into 0-value pixels
        
        if re_order:
            for m,mm in enumerate(sorted(np.unique(mask1))):
                if mask1.min() != 0:   ## if masks take up the entire space of the image / there is no background, then need to index from 1 instead of 0
                    m = m + 1
                mask1[mask1 == mm] = m

        return mask1

        

    ### This function calculates and writes the intesities, regionprops csv files. 
    def make_segmentation_measurements(self, 
                                       input_img_folder: Union[Path, str], 
                                       input_mask_folder: Union[Path, str], 
                                       output_intensities_folder: Union[Path, str, None] = None, 
                                       output_regions_folder: Union[Path, str, None] = None, 
                                       statistic: str= "mean",
                                       re_do: bool = False, 
                                       advanced_regionprops: bool = False, 
                                        ) -> None:                              
                                                                                # ****stein_derived (implements parts of 
                                                                                # the steinbock.measurement module scripts
                                                                                # # *** marks directly copied/derived lines)
        '''
        This method measures statistics and regionproperties from cell masks + images, and writes these as intensity 
        and regionprops csv files in an output folder. This output folder is structured such that a PalmettoBUG-style 
        analysis can easily be launched from it. 

        It is derived & and relies upon steinbock region_measurements functions (and through them, skimage). 

        Args:
            input_img_folder (Path, str): 
                the file path to the folder containing the images to be used for measuring 
                statistics / intensities

            input_mask_folder (Path string): 
                the file path to the folder contianing the masks to be used for measuring both 
                stats / intensities AND region properties (the region properties depend only on 
                the shape of the masks, not of the channels of the matching images)

                NOTE! -- the input_img_folder and input_mask_folder are presumed to have the same number of files & these files share 
                filenames / order. This is the default way that PalmettoBUG / isoSegDenoise exports these files, but be 
                careful to make sure to change this to be the case if your data does not match this pattern!

            output_intensities_folder (Path, string, None): 
                the file path to the folder to export the intensity csv files (these 
                csv files are effectively like fcs files with events ['Object'] each with intensity measurements / statistics for each channel). 
                If None, then the self.directory_object.intensities_dir will be used -- this depends on having set up an analysis using 
                the self.make_analysis_dirs(analysis_name) method beforehand

            output_intensities_folder (Path, string, None): 
                the file path to the folder contianing the regionprops csv files for each mask. 
                Like the intensities csv's these are structured with measurements for each object/event/cell, but are measurements like area, perimeter, etc.
                If None, then will use the self.directory_object.regionprops_dir -- this depends on calling the self.make_analysis_dirs(analysis_name) method beforehand

            statistic (string): 
                The statistic to report for each cell/object for each channel in the instensity csv files.
                One of ["mean","median","min","max","sum","std","var"]. Default is "mean", and you will rarely need any of the other options. 

            re_do (boolean): 
                Determines whether to check if each intensities/regionprops csv has already been generated (by matching the file name) 
                and ONLY export NEW csv files (if = True).
                Alternatively, export every csv file regardless of whether or not the export folder already contains identically named files (if = False). 

                Consider re_do = True if you are adding new files to a large existing project and don't want the old files to be redone or are 
                generating advanced regionprops (saves time), 

                otherwise re_do = False is frequently better as writing each file without advanced regionprops is not a very long process & ensures 
                consistency at this step. 

            advanced_regionprops (boolean): 
                If True, will calculate a few 'advanced' region properties, like --> [number of branch points, tortuosity, etc...]
                of the masks. This is a slow process so the default advanced_regionprops = False is preferred unless these additional region properties are greatly desired. 

                WARNING!! -- This option is currently broken / unreliable (algorithms & packages used to derive these regionproperties appear to make errors
                in determining branching.)

        Returns:
            None -- (its output is in writing to the disk, not returning a value)
        '''
        input_img_folder = str(input_img_folder)
        input_mask_folder = str(input_mask_folder)

        if output_intensities_folder is None:
            output_intensities_folder = self.directory_object.intensities_dir
        if output_regions_folder is None:
            output_regions_folder = self.directory_object.regionprops_dir

        output_intensities_folder = str(output_intensities_folder)
        output_regions_folder = str(output_regions_folder)

        if not os.path.exists(output_intensities_folder):
            os.mkdir(output_intensities_folder) 
        if not os.path.exists(output_regions_folder):
            os.mkdir(output_regions_folder) 

        dict_of_choices = {       # ***
            "sum": stein_unhook.IntensityAggregation.SUM,
            "min": stein_unhook.IntensityAggregation.MIN,
            "max": stein_unhook.IntensityAggregation.MAX,
            "mean": stein_unhook.IntensityAggregation.MEAN,
            "median": stein_unhook.IntensityAggregation.MEDIAN,
            "std": stein_unhook.IntensityAggregation.STD,
            "var": stein_unhook.IntensityAggregation.VAR,
            }
        img_files = sorted(Path(input_img_folder).rglob("[!.]*.tiff"))   # ***
        mask_files = sorted(Path(input_mask_folder).rglob("[!.]*.tiff"))  # ***
        ints_folder = sorted(Path(output_intensities_folder).rglob("[!.]*.csv"))   # ***
        regions_folder = sorted(Path(output_regions_folder).rglob("[!.]*.csv"))    # ***

        cleaned_img = []
        cleaned_mask = []
        for i in img_files:
            j = str(i).replace("\\","/")
            j = j[(j.rfind("/") + 1):j.rfind(".tiff")]
            cleaned_img.append(j)
        for ii in mask_files:
            jj = str(ii).replace("\\","/")
            jj = jj[(jj.rfind("/") + 1):jj.rfind(".tiff")]
            cleaned_mask.append(jj)
        shared_files = [i for i in cleaned_img if i in cleaned_mask]
        shared_masks = [i for i in cleaned_mask if i in cleaned_img]

        if (len(shared_files) == 0):
            if _in_gui:
                tk.messagebox.showwarning("Warning!", 
                    message = "None of the mask and image filenames matched! Cancelling regionproperty measurement.")
            else:
                print("None of the mask and image filenames matched! Cancelling regionproperty measurement.")
            return

        def filter_redo(dest_folder, shared_files):
            ''''''
            ints_files = []
            if dest_folder is not None:
                ints_files = [str(i).replace("\\","/") for i in dest_folder]
                ints_files = [i[(i.rfind("/") + 1):i.rfind(".csv")] for i in ints_files]
            img_files_int = []
            mask_files_int = []
            for i in shared_files:
                if i not in ints_files:
                    img_files_int.append(f'{input_img_folder}/{i}.tiff')
                    mask_files_int.append(f'{input_mask_folder}/{i}.tiff')
            return img_files_int, mask_files_int

        def write_csvs(img_files, generator, out_directory, csv_type, input_mask_folder = input_mask_folder):
            ''''''
            for _ in img_files:
                img_file, mask_file, csv = next(generator)
                if csv_type == "regions":
                    csv['image_area'] = tf.imread(mask_file).shape[0] * tf.imread(mask_file).shape[1]
                    csv['mask_folder'] = input_mask_folder
                right_index1 = str(mask_file).rfind('/')
                right_index2 = str(mask_file).rfind('\\')
                right_index = np.max([right_index1,right_index2])
                left_index = str(mask_file).rfind('.tiff')
                file_name = str(mask_file)[right_index+1:left_index]
                if (len(csv[csv.columns[0]]) != 0):           #### This means there are no cell masks in this file! 
                    csv.to_csv(("".join([out_directory, '/', file_name, ".csv"])),index = True)
                    print(f"{file_name} {csv_type} csv has been written!") 
                else:
                    if _in_gui:
                        warning_window(f"""{file_name} has no cell masks in it! 
                            Re-run segmentation or delete image & its mask / csv's from the analysis! 
                            \n conversion to Analysis will fail in the creation of a 0 event fcs""")
                    print(f"""{file_name} has no cell masks in it! 
                        Re-run segmentation or delete image & its mask / csv's from the analysis! 
                        \n conversion to Analysis will fail in the creation of a 0 event fcs""")
        
        if re_do is False:
            img_files_int, mask_files_int = filter_redo(ints_folder, shared_files)
            img_files_reg, mask_files_reg = filter_redo(regions_folder, shared_masks)
            if (len(img_files_int) == 0) and (len(img_files_reg) == 0):
                if _in_gui:     
                    tk.messagebox.showwarning("Warning!", 
                        message = "All images have intensity and region files written! Did you intend to redo these measurements?")  
                else:
                    print("All images have intensity and region files written! Did you intend to redo these measurements?")  
                return
            
            intensity_gen = stein_unhook.try_measure_intensities_from_disk(img_files_int,           # *** 
                                        mask_files_int, 
                                        self.panel[self.panel['keep'] == 1]['name'], 
                                        dict_of_choices[statistic])                     
            regionprops_gen = stein_unhook.try_measure_regionprops_from_disk(img_files_reg,         # ***
                                                                             mask_files_reg,["area", 
                                                                                    "perimeter",              
                                                                                    "centroid",
                                                                                    "axis_major_length",
                                                                                    "axis_minor_length",
                                                                                    "eccentricity"])
         
            write_csvs(img_files_int, intensity_gen, output_intensities_folder, csv_type = "intensities")
            write_csvs(img_files_reg, regionprops_gen, output_regions_folder, csv_type = "regions" )

        else:
            img_files_int, mask_files_int = filter_redo(None, shared_files)
            intensity_gen = stein_unhook.try_measure_intensities_from_disk(img_files_int,                      # *** 
                                                        mask_files_int, 
                                                        self.panel[self.panel['keep'] == 1]['name'], 
                                                        dict_of_choices[statistic])  
            regionprops_gen = stein_unhook.try_measure_regionprops_from_disk(img_files_int,                    # ***
                                                            mask_files_int,["area", 
                                                                        "perimeter",
                                                                        "centroid",
                                                                        "axis_major_length",
                                                                        "axis_minor_length",
                                                                        "eccentricity"])
            write_csvs(img_files_int, intensity_gen, output_intensities_folder, csv_type = "intensities")
            write_csvs(mask_files_int, regionprops_gen, output_regions_folder, csv_type = "regions" )
                    
        if advanced_regionprops is True:
            print("Beginning advanced regionprop calculations -- this could take some time")
            self._advanced_regionprops(input_mask_folder, output_regions_folder)

    def _advanced_regionprops(self, 
                              input_mask_folder: Union[Path, str], 
                              output_regions_folder: Union[Path, str],
                              ) -> None:   
        '''
        Helper function for self.make_segmentation_measurements(). Calculates 'advanced' region properties: [n_slab, n_branch, tortuosity, cycles]

        WARNING!! -- Currently not accurate / functional (inaccurate, additional branches in simple tests). Error may propagate from NAVis library
        '''
        try:
            import navis as nv #type: ignore
        except Exception:
            print("Advanced Regionprops requires the navis package to be installed! Warning -- test of navis branching algorithm in simple cases had errors (note on 1/13/25)")
            return
        input_mask_folder = str(input_mask_folder)
        output_regions_folder = str(output_regions_folder)

        mask_files = ["".join([input_mask_folder,"/",i]) for i in sorted(os.listdir(input_mask_folder)) if i.lower().find(".tif") != -1]
        regionprops_files = ["".join([output_regions_folder,"/",i]) for i in sorted(os.listdir(output_regions_folder)) if i.lower().find(".csv") != -1]
        for i,ii in zip(mask_files, regionprops_files):
            image = tf.imread(i).astype('int')
            n_slab_list = []
            n_branch_list = []
            tortuosity_list = []
            cycle_list = []
            for kk,k in enumerate(skimage.measure.regionprops(image)):
                to_navis = image[k.bbox[0]:k.bbox[2],k.bbox[1]:k.bbox[3]]
                skeleton = skimage.morphology.skeletonize(to_navis).astype('int')
                to_navis = np.pad(skeleton[:,:,np.newaxis], 1)
                mesh = skimage.measure.marching_cubes(to_navis, spacing = (1.0, 1.0, 0.01))
                verts, faces, _, _ = mesh 
                try:
                    verts[:,2] = 0
                    meshy_neury = nv.MeshNeuron((verts, faces))   # type: ignore     # TODO: reactivate this function once fixed
                    tree_neuree = meshy_neury.skeletonize()   
                    n_slab_list.append((tree_neuree.nodes['type'] == 'slab').sum())
                    n_branch_list.append((tree_neuree.nodes['type'] == 'branch').sum())
                    cycles = tree_neuree.cycles
                    if cycles is not None:
                        cycle_list.append(cycles)
                    else:
                        cycle_list.append(0) 
                except Exception:
                    n_slab_list.append(0)
                    n_branch_list.append(0)
                    cycle_list.append(0)
                try:
                    tortuosity_list.append(nv.tortuosity(tree_neuree))     # type: ignore    # TODO: reactivate this function once fixed
                except Exception:
                    tortuosity_list.append(1)
            df = pd.read_csv(ii)
            df['n_slab'] = n_slab_list
            df['n_branch_list'] = n_branch_list
            df['tortuosity'] = tortuosity_list
            df['cycles'] = cycle_list
            df.to_csv(ii, index = False)    

    def to_analysis(self, Analysis_tab = None, metadata_from_save: bool = False) -> None:
        '''
        This function prepares / sets up an Analysis folder, by converting intensity csv files to fcs files and generating preliminary / semi-empty metadata / panel 
        pandas dataframes, which require editing before writing to the disk at the newly prepared self.Analysis_panel_dir and self.metadata_dir filepaths.

        Depends on self.make_analysis_dirs(analysis_name) being called first to set up the directory structure and direct the discovery
        of the intensity files which will be used to generate FCS files & the intial panel/metadata dataframes.

        DOES NOT export the panel / metadata files to the disk, but returns them, along with the file paths where they are expected to
        be wrtten to, in the following order: (panel dataframe, metadata dataframe, panel path, metadata path)

        Args:
            Analysis_tab / metadata_from_save (Only for use inside GUI): 
                IGNORED outside of GUI. in the GUI they assist in coordinating the widgets & choosing to load the Analysis_panel/metadata files 
                from the directory_object.Analyses_dir (if someone makes a second analysis in one project, instead of requiring a fresh panel/metadata 
                set up each time a new analysis is made). 
        '''
        #print(_in_gui)
        if not _in_gui:
            self._intense_to_fcs()
            if ((not os.path.exists(self.directory_object.Analyses_dir + "/Analysis_panel.csv")) 
                or (not os.path.exists(self.directory_object.Analyses_dir + "/metadata.csv"))):
                print("""Panel / Metadata could not be loaded from save (in Analyses directory) -- generating these from scratch""")
                panel_file = self._initial_Analysis_panel()
                metadata = self._initial_metadata_file()
            else:
                panel_file = pd.read_csv(self.directory_object.Analyses_dir + "/Analysis_panel.csv")
                metadata = pd.read_csv(self.directory_object.Analyses_dir + "/metadata.csv")
            self.Analysis_panel_dir = self.directory_object.Analysis_internal_dir + "/Analysis_panel.csv"
            self.metadata_dir = self.directory_object.Analysis_internal_dir + "/metadata.csv"
            return panel_file, metadata, self.Analysis_panel_dir, self.metadata_dir
    
        elif _in_gui:
            self._intense_to_fcs()
            Analysis_tab.analysiswidg.setup_dir_disp(self.directory_object.analysis_dir + "/main") 
            Analysis_tab.master.Spatial.widgets.setup_dir_disp(self.directory_object.analysis_dir + "/main")
            if metadata_from_save is True: 
                if ((not os.path.exists(self.directory_object.Analyses_dir + "/Analysis_panel.csv")) 
                or (not os.path.exists(self.directory_object.Analyses_dir + "/metadata.csv"))):
                    tk.messagebox.showwarning("Warning!",
                        message = """Loading Panel / Metadata from save option checked -- 
                                    but one or both of those files is not present in the /Analyses directory!""")
                self.Analysis_panel = pd.read_csv(self.directory_object.Analyses_dir + "/Analysis_panel.csv")
                self.metadata = pd.read_csv(self.directory_object.Analyses_dir + "/metadata.csv")

                analysis_logger = Analysis_logger(self.directory_object.Analysis_internal_dir).return_log()
                table_launcher = TableLaunchAnalysis(1, 1, 
                                                    self.directory_object.Analysis_internal_dir, 
                                                    self.Analysis_panel, "Analysis_panel.csv", 
                                                    self, Analysis_tab.master, 
                                                    self.directory_object.Analyses_dir, 
                                                    favor_table = True, 
                                                    logger = analysis_logger)
                table_launcher.add_table(1, 1, 
                                         self.directory_object.Analysis_internal_dir, 
                                         self.metadata, "metadata",
                                         favor_table = True)
            else:
                self._initial_Analysis_panel()
                analysis_logger = Analysis_logger(self.directory_object.Analysis_internal_dir).return_log()
                table_launcher = TableLaunchAnalysis(1, 1, 
                                                     self.directory_object.Analysis_internal_dir, 
                                                     self.Analysis_panel, 
                                                    "Analysis_panel.csv", 
                                                    self, 
                                                    Analysis_tab.master, 
                                                    self.directory_object.Analyses_dir, 
                                                    logger = analysis_logger)
                self._initial_metadata_file()
                table_launcher.add_table(1, 1, 
                                         self.directory_object.Analysis_internal_dir, 
                                         self.metadata, "metadata")

    def _intense_to_fcs(self, 
                        input_intensity_directory: Union[Path, str, None] = None, 
                        ouput_fcs_folder: Union[Path, str, None] = None,
                        ) -> None:
        '''
        Helper method for self.to_Analysis --> writes .fcs files from .csv files in the intensities folder.

        When called with default arguments, depends on self.make_analysis_dirs(analysis_name), however the input and output folders
        can be specified separately to allow its use in any context.

        Args:
            input_intensity_directory (Path, string, None): 
                the path to a folder where the intensity csv files are (exported by self.make_segmentation_measurements)
                If None, then a default path is presumed which requires a prior execution of the self.make_analysis_dirs(analysis_name) method

            ouput_fcs_folder (Path, string, None): 
                the path to a folder where the FCS files will be written to, with the same behaviour as (input_intensity_directory) - if None, a 
                default path self.directory_object.fcs_dir is presumed, etc.
        '''
        if input_intensity_directory is None:                               
            input_intensity_directory = self.directory_object.intensities_dir
        if ouput_fcs_folder is None:
            ouput_fcs_folder = self.directory_object.fcs_dir 
        input_intensity_directory = str(input_intensity_directory)
        ouput_fcs_folder = str(ouput_fcs_folder)
        if not os.path.exists(ouput_fcs_folder):
            os.mkdir(ouput_fcs_folder)

        for i in sorted(os.listdir(input_intensity_directory)):
            if i.lower().find(".csv") != -1:
                pd_df = pd.read_csv("".join([input_intensity_directory, "/", i]))
                fcs_df = DataFrame(pd_df, columns = pd_df.columns)
                filename = i[:i.rfind('.')]
                fcs_df.to_fcs("".join([ouput_fcs_folder, "/", filename, ".fcs"]))

    def _initial_Analysis_panel(self) -> pd.DataFrame:
        ''' 
        Helper method for self.to_Analysis generates an initial Analysis panel (marker_class column blank)
        '''
        self.Analysis_panel = self.panel.drop(['segmentation','channel'], axis = 1)
        self.Analysis_panel = self.Analysis_panel[self.Analysis_panel['keep'] == 1]
        self.Analysis_panel = self.Analysis_panel.reset_index().drop(['index','keep'], axis = 1)
        self.Analysis_panel['fcs_colname'] = self.Analysis_panel['name']
        self.Analysis_panel['antigen'] = self.Analysis_panel['name']
        self.Analysis_panel = self.Analysis_panel.drop('name', axis = 1)
        self.Analysis_panel['marker_class'] = 'none'  
        return self.Analysis_panel
    
    def _initial_metadata_file(self) -> pd.DataFrame:
        '''
        Helper method for self.to_Analysis --> generate the initial metadata file (patient and condition columns blank)
        '''
        file_list = [i for i in sorted(os.listdir(self.directory_object.fcs_dir)) if i.lower().find(".fcs") != -1]
        metadata = pd.DataFrame()
        metadata['file_name']  = file_list
        metadata['sample_id'] = metadata.reset_index()['index']
        metadata['patient_id'] = '' 
        metadata['condition'] = 'treatment vs. control'   
        self.metadata = metadata
        return metadata
    
def setup_for_FCS(directory):
    '''
    This sets up a folder for single-cell analysis using the CATALYST-derived module of PalmettoBUG. 

    Can be used for a solution-mode experiment (direct from FCS files) or as part of the set up when transitioning from image processing to single-cell analysis.

    Args:
        directory (str):
            The directory to set up from single-cell analysis

    Returns:
        Analysis_panel (a pandas dataframe of the initially generated Analysis_panel file, needs the marker_class column to be filled in by the user)

        Analysis_panel_dir (a string, the path to where the Analysis_panel should be saved on the disk once it has been completed by the user)

        metadata (a pandas dataframe of the initially generated metadata file, needs the patient_id and condition columns to be filled in by the user)

        metadata_dir (a string, the path to where the metadata should be saved on the disk once it has been completed by the user)
    '''
    directory_name = directory[:directory.rfind("/")]

    directory_object = DirSetup(directory, kind = "Analysis")
    directory_object.make_analysis_dirs(directory_name)
    Analysis_panel_dir = directory_object.Analysis_internal_dir + "/Analysis_panel.csv"
    metadata_dir = directory_object.Analysis_internal_dir + "/metadata.csv"
    fcs_files = [i for i in sorted(os.listdir(directory_object.fcs_dir)) if i.lower().find(".fcs") != -1]
    try:
        Analysis_panel = pd.read_csv(Analysis_panel_dir)
    except FileNotFoundError:
        warnings.filterwarnings("ignore", message = "The default channel names")
        _, dataframe1 = fcsparser.parse(directory_object.fcs_dir + "/" + fcs_files[0])
        warnings.filterwarnings("default", message = "The default channel names")
        Analysis_panel = pd.DataFrame()
        Analysis_panel['fcs_colnames'] = dataframe1.columns
        Analysis_panel['antigen'] = dataframe1.columns
        Analysis_panel['marker_class'] = "none"
        try:
            Analysis_panel = Analysis_panel.drop("Object", axis = 0) 
        except KeyError:
            pass
        Analysis_panel.to_csv(Analysis_panel_dir, index = False) 

    try:
        metadata = pd.read_csv(metadata_dir)
        
    except FileNotFoundError:
        metadata = pd.DataFrame()
        metadata['file_name']  = fcs_files
        metadata['sample_id'] = metadata.reset_index()['index']
        metadata['patient_id'] = 'na'      # manually set later
        metadata['condition'] = 'treatment vs. control'       # manually set later

    return Analysis_panel, Analysis_panel_dir, metadata, metadata_dir

class direct_to_Analysis(ImageAnalysis):
    ''' GUI only '''
    def __init__(self, 
                 master, 
                 directory: str, 
                 metadata_from_save: bool = False):      
        ## this is essentially its own, customized directory then panel/metadata setup (particularly the panel setup is customized)
        self.master = master
        directory_name = directory  #[:directory.rfind("/")]
        self.directory_object = DirSetup(directory, kind = "Analysis")
        self.directory_object.make_analysis_dirs(directory_name)
        if metadata_from_save is True:                           ## the loader from the common / saved metadata
            self.Analysis_panel_dir = self.directory_object.Analyses_dir + "/Analysis_panel.csv"
            self.metadata_dir = self.directory_object.Analyses_dir + "/metadata.csv"
        else:
            self.Analysis_panel_dir = self.directory_object.Analysis_internal_dir + "/Analysis_panel.csv"
            self.metadata_dir = self.directory_object.Analysis_internal_dir + "/metadata.csv"

        # read the first .fcs file to get the column names, then write panel file:
        # this has to be done differently than the prior ImageAnalysis class, as it is starting halfway 
        #   and without a pre-existing (steinbock-style) panel file
        try:
            self.Analysis_panel = pd.read_csv(self.Analysis_panel_dir)
        except FileNotFoundError:
            fcs_files = [i for i in sorted(os.listdir(self.directory_object.fcs_dir)) if i.lower().find(".fcs") != -1]
            warnings.filterwarnings("ignore", message = "The default channel names")
            _, dataframe1 = fcsparser.parse(self.directory_object.fcs_dir + "/" + fcs_files[0])
            warnings.filterwarnings("default", message = "The default channel names")
            self.Analysis_panel = pd.DataFrame()
            self.Analysis_panel['fcs_colnames'] = dataframe1.columns
            self.Analysis_panel['antigen'] = dataframe1.columns
            self.Analysis_panel['marker_class'] = "none"
            try:
                self.Analysis_panel = self.Analysis_panel.drop("Object", axis = 0) 
            except KeyError:
                pass
            self.Analysis_panel.to_csv(self.Analysis_panel_dir, index = False) 

        # now do the metadata initialization and table launches (shared with the main pathway)
        analysis_logger = Analysis_logger(self.directory_object.Analysis_internal_dir).return_log()
        table_launcher = TableLaunchAnalysis(1, 1,  
                                             self.directory_object.Analysis_internal_dir, 
                                             self.Analysis_panel, 
                                             "Analysis_panel.csv", 
                                             self, 
                                             self.master, 
                                             alt_dir = self.directory_object.Analyses_dir, 
                                             logger = analysis_logger)
        fcs_files  = [i for i in sorted(os.listdir(self.directory_object.fcs_dir)) if i.lower().find(".fcs") != -1]
        try:
            self.metadata = pd.read_csv(self.metadata_dir)
            
        except FileNotFoundError:
            self.metadata = pd.DataFrame()
            self.metadata['file_name']  = fcs_files
            self.metadata['sample_id'] = self.metadata.reset_index()['index']
            self.metadata['patient_id'] = 'na'      # manually set later
            self.metadata['condition'] = 'treatment vs. control'       # manually set later

        table_launcher.add_table(1, 1, 
                                 self.directory_object.main, 
                                 self.metadata, 
                                 "metadata")


class TableLaunchAnalysis(TableLaunch):
    ''' GUI only '''
    def __init__(self, width: int, 
                 height: int, 
                 directory: str, 
                 dataframe: pd.DataFrame,
                 table_type: str, 
                 experiment: ImageAnalysis, 
                 tab, 
                 alt_dir: Union[str, None] = None,
                 favor_table: bool = False, 
                 logger = None):
        ''''''
        super().__init__(width, 
                         height, 
                         directory, 
                         dataframe, 
                         table_type, 
                         experiment, 
                         favor_table = favor_table, 
                         logger = logger)
        self.geometry("+0+0")
        self.tab = tab
        self.alt_dir = alt_dir
        self.Analysis_internal_dir = directory 

        self.save_meta_panel = ctk.CTkCheckBox(master = self, 
            text = """Check to save metadata/Analysis_panel to \n Analyses directory \n (can be re-used when making another analysis)""", 
            onvalue = True, 
            offvalue = False)
        self.save_meta_panel.grid(column = 1, row = 2, pady = 15)

    def accept_and_return(self, experiment: ImageAnalysis) -> None:     
        #### overwrite the accept & return function from the parent to also populate the Analysis widgets
        save_or_not = self.save_meta_panel.get()
        for i in self.table_list:
            i.recover_input()
            if i.type == "Analysis_panel":
                try:
                    experiment.Analysis_panel = i.table_dataframe
                except AttributeError:
                    pass
                self.Analysis_panel_write(i.table_dataframe)
                if save_or_not is True:
                    self.Analysis_panel_write(i.table_dataframe, alt_directory = self.alt_dir)
            if i.type == "metadata":
                try:
                    experiment.metadata = i.table_dataframe    #### this is for the whole class analysis -- it is kind of awkward though
                except AttributeError:
                    pass
                self.metadata_write(i.table_dataframe)
                if save_or_not is True:
                    self.metadata_write(i.table_dataframe, alt_directory = self.alt_dir)
        if self.tab is not None:
            self.tab.py_exploratory.analysiswidg.initialize_experiment_and_buttons(experiment.directory_object.Analysis_internal_dir)
        self.after(200, self.destroy())



'''
This module contains the back-end class (WholeClassAnalysis) that handles the analysis of pixel classes as-a-whole. It is accessed in the GUI
through the second half of the third tab of the program (use pixel classifier).
It is also available through the public (non-GUI) API of PalmettoBUG. 
'''
## License / derivation info
# PalmettoBUG as a whole, and therefore also this script, is licensed under the GPL3 license
#
# This class is largely based of off the Analysis class of PalmettoBUG (see Analysis.py file), which in turn was heavily inspired / derived from CATALYST
# However, this class shares very limited similarity with CATALYST besides the panel / metadata files.
# See Assets / Other_License_Details.txt for more information on 3rd-party sources of code / ideas in this package.

import os
from pathlib import Path
from typing import Union
import warnings

import scipy
import anndata as ann
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from .._vendor import sigfig

warnings.filterwarnings("ignore", message = "Transforming to str index")   ## anndata implicit modification warning that is not necessary
warnings.filterwarnings("ignore", message = "Observation names are not unique")  ## anndata UserWarning that is not necessary

__all__ = ["WholeClassAnalysis"]

class WholeClassAnalysis:     
    '''
    This class handles the whole-class Analysis, where pixel regions are treated as if they are cell segmentation masks

    It has limited options compared to the standard experiment class that handles true single cell data. This class only has a few plot options 
    and a single statistics option, and no batch correction, dropping of samples, or scaling

    Args:
        directory (string or Path): the path to a folder containing /intensities and /regionprops subfolders,
                which would have been produced by running region measurements on the pixel classification maps
                generated by a pixel classifier. 

        classifier_df (pandas dataframe): the biological_labels.csv exported from the pixel classifier whose
                output is being used. Contains "labels", "class", and/or "merge" columns,
                which help associate region numbers in the images / regionpros & intensity csvs
                with the biological labels in the classifier. 

        metadata (pandas dataframe): analogous to the metadata csv file in a standard, single-cell analysis
                Contains the same, file_name, sample_id, patient_id, condition columns

        Analysis_panel (pandas dataframe): analogous to the Analysis_panel csv file in a standard, single-cell analysis.
                For example, contains columns for antigen and marker_class.

        class_type (string): one of -- "premerge", "merged" -- whether the outputs of the classifier are before or after
                merging (relevant for what column in classifier_df is used as the class, "merging" or "class" )

    Key Attributes:
        data (anndata.AnnData): the data, with data.X being a numpy array containing the channel information per "event" (per class per image)
                data.obs being derived from the inputted metadata and data.var being derived from the Analysis_panel

        class_labels (pandas DataFrame): this is the inputted classifier_df, which associates the class numbers with biological labels

        directory (str): The path to the folder where the analysis is to be initialized. Used to set up directories 
                (such as save_dir, data_table_dir), to export some files (input_tables_to_csv) and to find the expected
                intensities / regionprops csv files when loading the data. 

        save_dir (str): The path to where plots are saved by this class (when filename is provided in plotting functions)

        data_table_dir (str): The path to where data tables are saved by this class (when filename is provided to methods that produce
                dataframes such as statistics / exports) 
    '''
    def __init__(self, 
                 directory: Union[Path, str], 
                 classifier_df: pd.DataFrame, 
                 metadata: pd.DataFrame, 
                 Analysis_panel: pd.DataFrame, 
                 csv: Union[pd.DataFrame, None] = None,
                 ) -> None:
        '''
        '''
        self.directory = str(directory)
        self.directory  = self.directory.replace("\\" , "/")

        classifier_df['class'] = classifier_df['class'].astype('int')
        classifier_df['merging'] = classifier_df['merging'].astype('int')
        self.class_labels = classifier_df
        self._metadata = metadata
        try:
            old_index = Analysis_panel.index.copy()
            Analysis_panel.index = Analysis_panel['antigen']
            Analysis_panel = Analysis_panel.drop('Object', axis = 0)
            Analysis_panel.index = old_index    ## if present, remove the Object column from the panel
        except Exception:
            pass
        self._panel = Analysis_panel

        self.save_dir = self.directory + "/python_plots"
        self.data_table_dir = self.directory + "/Data_tables"
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

        if not os.path.exists(self.directory + "/Data_tables"):
            os.mkdir(self.directory + "/Data_tables")
        
        self.percent_areas = None
        if csv is None:
            self._load()
        else:
            self._load(csv = csv)

    def _load(self, csv: Union[pd.DataFrame, None] = None, arcsinh_cofactor = 5) -> None:
        '''Helper to the __init__ method: performs the loading and shaping of data during the initial load.'''
        metadata = self._metadata
        panel = self._panel
        panel.index = panel['antigen']

        if csv is not None:
            marker_class_included = False
            marker_class = csv.copy().iloc[-1,:]
            if np.array(marker_class == "na").sum() != 0: 
                marker_class_dict_rev = {"0.0" : 'none', "1.0" : 'type', "2.0" : ' state'}
                marker_class = marker_class[marker_class != "na"].astype('str').replace(marker_class_dict_rev)
                csv = csv.iloc[:-1, :]
                marker_class_included = True

            metadata_columns = ["class", "sample_id", "patient_id", "condition","file_name","Object"]
            to_obs = csv[[i for i in csv.columns if i in metadata_columns]].astype('str')
            not_obs = csv[[i for i in csv.columns if i not in metadata_columns]].astype('float')
            self._metadata = to_obs
            self._panel = pd.DataFrame()
            self._panel['antigen'] = list(not_obs.columns)
            self._panel['name'] = list(not_obs.columns)
            self._panel.index = self._panel['antigen']
            if marker_class_included is True:
                self._panel['marker_class'] = list(marker_class) + [1] ## Areas will always be assumed type
            else:
                self._panel['marker_class'] = 'type'
                print("All antigens from the csv have been set to 'type' in the panel file! \n"
                      "Open the Analysis_panel.csv file, edit, & reload the experiment to change this!")

            if len(to_obs["Object"].unique()) == len(self.class_labels['labels'].unique()):
                self.class_type = "premerge"
            elif len(to_obs["Object"].unique()) == (len(self.class_labels['labels'].unique()) - 1):
                self.class_type = "merged"
                self.class_labels = self.class_labels[self.class_labels['labels'] != 'background']

            self.data = ann.AnnData(X = np.array(not_obs), obs = to_obs, var = self._panel)
            self._object_column = to_obs['Object']
        
        else:
            csv_directory = self.directory  + "/intensities"
            csv_dir_names = [i for i in sorted(os.listdir(csv_directory)) if i.lower().find(".csv") != -1]

            if len(csv_dir_names) != len(self._metadata):
                print("Metadata file_name column and number of csv files in analysis do not match --\n" 
                      "only keeping data present in both for analysis!")
                new_fcs_filenames = []
                truth_array = np.zeros(len(self._metadata)).astype('bool')
                for i in list(self._metadata['file_name']):

                    if i in csv_dir_names:
                        new_fcs_filenames.append(i)
                        truth_array = truth_array  + np.array(self._metadata['file_name'] == i).astype('bool')

                self._metadata = self._metadata[truth_array]
                csv_dir_names = new_fcs_filenames

            csv_path_list = ["".join([csv_directory,"/",i]) for i in csv_dir_names]

            intensities = pd.DataFrame()
            length_of_images = [0]
            length_of_images2 = []
            tally = 0
            for i in csv_path_list:
                fcs_read_in = pd.read_csv(i)
                tally = tally + len(fcs_read_in.index)
                length_of_images.append(tally)
                length_of_images2.append(len(fcs_read_in.index))
                intensities = pd.concat([intensities, fcs_read_in], axis = 0)

            self._object_column = intensities['Object']
            intensities = intensities.drop("Object", axis = 1)

            if len(self._object_column.unique()) == len(self.class_labels['labels'].unique()):
                self.class_type = "premerge"
            elif len(self._object_column.unique()) == (len(self.class_labels['labels'].unique()) - 1):
                self.class_type = "merged"
                self.class_labels = self.class_labels[self.class_labels['labels'] != 'background']
            else:
                raise Exception("The number of classes in the data and the number of classes in the biological_labels.csv do not match! \n\n"
                                "(They are neither equal innumber nor does the biological_labels.csv has +1 class [background])")

            if self.class_type == "merged":
                zip_dict = {}
                for i,ii in zip(self.class_labels['merging'],self.class_labels['labels']):
                    zip_dict[i] = ii
            else:
                zip_dict = {}
                for i,ii in zip(self.class_labels['class'],self.class_labels['labels']):
                    zip_dict[i] = ii
            if arcsinh_cofactor > 0:
                exprs = pd.DataFrame(np.arcsinh(intensities / arcsinh_cofactor))
            else:
                exprs = pd.DataFrame(intensities)
            exprs.columns = panel["antigen"]

            metadata_long = pd.DataFrame()
            sample_id_array = np.zeros([0])
            counter = 0
            for i,ii in zip(length_of_images[:-1], length_of_images[1:]):
                slicer = np.full(shape = [ii - i], fill_value = counter)
                sample_id_array = np.append(sample_id_array,slicer)
                counter += 1
                
            metadata_long['sample_id'] = sample_id_array.astype('int')

            meta_file_dict = {}
            meta_patient_dict = {}
            meta_condition_dict = {}
            for i,ii in enumerate(metadata["sample_id"]):
                meta_file_dict[ii]= list(metadata.iloc[i])[0]
                meta_patient_dict[ii]= list(metadata.iloc[i])[2]
                meta_condition_dict[ii]= list(metadata.iloc[i])[3]

            filenames = metadata_long.replace(meta_file_dict)
            patient_ids = metadata_long.replace(meta_patient_dict)
            conditions = metadata_long.replace(meta_condition_dict)

            metadata_long['file_name'] = filenames
            metadata_long['patient_id'] = patient_ids
            metadata_long['condition'] = conditions
            metadata_long['Object'] = list((self._object_column).astype('int'))
            metadata_long['class'] = metadata_long["Object"].astype('int').replace(zip_dict).astype('str')
            metadata_long.index = exprs.index

            self._metadata["number_of_cells"] = length_of_images2

            self.data = ann.AnnData(X = exprs, var = panel, obs = metadata_long)
            self.data.obs["sample_id"] =  self.data.obs["sample_id"].astype("category")
            self.data.obs["patient_id"] =  self.data.obs["patient_id"].astype("category")
            self.data.obs["condition"] =  self.data.obs["condition"].astype("category")
            self.data.obs = self.data.obs.reset_index().drop("index", axis = 1)

            self.plot_percent_areas(filename = None, calculate_only = True)   ## now this adds the "areas" column to the metadata and 
                                                                                # saves the self.percent_areas table ready for plotting

            self.data.uns['counts'] = np.array(intensities)   ## this will be the storage place for the pre-arcsinh transformed counts data 
                                                                        # (never used by GUI)

    def input_tables_to_csv(self) -> None:
        '''Allows the saving of the primary csv files within this class to the disk inside the self.directory folder'''
        self.class_labels.to_csv(self.directory + "/classifier_labels.csv", index = False)
        self._metadata.to_csv(self.directory + "/metadata.csv", index = False)
        self._panel.to_csv(self.directory + "/Analysis_panel.csv", index = False)

    def plot_percent_areas(self, filename: Union[str, None] = None, calculate_only: bool = False) -> plt.figure:
        ''' 
        Plots a boxplots of percent class in each image, showing and comparing the distributions between conditions
        
        Returns the plot as a matplotlib figure
        
        Args:
            filename (str or None):
                If filename is specified, this will export the plot as a PNG file to self.save_dir/{filename}.png

            calculate_only (bool):
                Default == False. If True (& self.percent_areas == None), this method will not return anything, 
                but instead will perform the calculation of %pixel class in each ROI. This calculation will be 
                saved to self.percent_areas, where it can easily be plotted by this function later. 
                This is implemented to save time by meaning the calculations only have to be done once

        Returns:
            matplotlib.pyplot figure or None (returns None only if calculate_only == True, and no 
            prior calculation of the % areas has been done)
        '''
        if self.percent_areas is None:
            whole_regionprops_dir = self.directory  + "/regionprops"
            data_whole = self.data.copy()
            data_whole_df = pd.DataFrame(data_whole.X, columns = data_whole.var.index, index  = data_whole.obs.index)
            data_whole_df["sample_id"] = data_whole.obs["sample_id"]
            data_whole_df["patient_id"] = data_whole.obs["patient_id"]
            data_whole_df["condition"] = data_whole.obs["condition"]
            data_whole_df['Object'] = list((self._object_column).astype('int'))
            data_whole_df['class'] = data_whole.obs["class"]

            ind_var_to_sample_id = {}
            for i,ii in zip(data_whole_df['sample_id'].astype('str'), data_whole_df["condition"]):
                ind_var_to_sample_id[i] = ii 

            data_whole_means = data_whole_df.groupby(["class","sample_id"], observed = True).mean(numeric_only = True).reset_index()
            data_whole_means['condition'] = data_whole_means['sample_id'].astype('str').replace(ind_var_to_sample_id)
            data_whole_means = pd.melt(data_whole_means, id_vars = ['condition',"class","sample_id"])
            grand_csv = pd.DataFrame()
            regionprop_file_list = [i for i in sorted(os.listdir(whole_regionprops_dir)) if i.lower().find(".csv") != -1]
            for i in regionprop_file_list:
                csv = pd.read_csv(whole_regionprops_dir + "/" + i)
                grand_csv = pd.concat([grand_csv,csv], axis = 0)
                
            grand_csv['class'] = list(data_whole_df['class'])
            grand_csv['sample_id'] = list(data_whole_df['sample_id'])
            grand_csv['condition'] = list(data_whole_df['condition'])
            grand_csv['area'] = grand_csv['area'] / grand_csv['image_area']      ## Percentile of total image area
            
            ind_var_to_sample_id = {}
            for i,ii in zip(grand_csv['sample_id'], grand_csv["condition"]):
                ind_var_to_sample_id[i] = ii 

            percent_areas = grand_csv[['area','class','sample_id']].groupby(['class','sample_id'], observed = False).mean() * 100
            percent_areas = percent_areas.reset_index()
            percent_areas['sample_class'] = [i + ii  for i, ii in zip(percent_areas['class'].astype('str'),percent_areas['sample_id'].astype('str'))]
            percent_dict = {}
            for i,ii in zip(percent_areas['area'],percent_areas['sample_class']):
                percent_dict[ii] = float(i)
            percent_areas['condition'] = percent_areas['sample_id'].replace(ind_var_to_sample_id)
            data_whole_df['sample_class'] = [i + ii  for i, ii in zip(data_whole_df['class'].astype('str'),data_whole_df['sample_id'].astype('str'))]
            data_whole_df['areas******'] = data_whole_df['sample_class'].astype('str').replace(percent_dict).astype('float')
            data_whole_df = data_whole_df.drop('sample_class', axis = 1)

            if calculate_only:
                self.data.obs['areas'] = list(data_whole_df['areas******'])
                self.percent_areas = percent_areas
                return
        else:
            percent_areas = self.percent_areas

        figure = plt.figure()
        ax = plt.gca()
        sns.boxplot(percent_areas, x = 'class', y = 'area', hue = 'condition', ax = ax)
        ax.set_title("% Area of each class type in each condition:")
        if filename is not None:
            figure.savefig(self.save_dir + f"/{filename}.png", bbox_inches = "tight") 
        plt.close()
        return figure
    
    def plot_distribution_exprs(self, 
                                unique_class: Union[str, int], 
                                plot_type: str,
                                marker_class: str = "All",
                                filename: Union[str, None] = None,
                                ) -> sns.FacetGrid:
        '''
        Plots a Bar or Violin plot from the distribution of marker expression / %class in each sample_id, comparing conditions
        
        Args: 
            unique_class (string or integer):
                Indicates which pixel class to plot antigen expressions for

            plot_type (string):
                'Violin' or 'Bar' -- determines what kind of plot is created

            marker_class (string):
                'All', 'type', 'state', or 'none' (or any other marker_class in self.data.var['marker_class']). Determines which antigens are used in the plot
                By default, every antigen, regardless of marker_class is used ('All'). 

            filename:
                If specified, this funciton will additionally export the plot as a PNG file to self.save_dir/{filename}.png

        Returns:
            the plot as a seaborn FacetGrid (FacetGrid.figure --> a matplotlib figure)
        '''
        plot_type_dict = {"Bar":sns.barplot, "Violin":sns.violinplot}
        plot_type_fx = plot_type_dict[plot_type]
        text_size = 9
        palette = 'tab20'
        legend_loc = [1.1,0.75]
        col_wrap = 4
        height = 12
        width = 7

        data_whole = self.data.copy()
        panel = self._panel.copy()
        if marker_class != "All":
            data_whole = data_whole[:, panel["marker_class"] == marker_class]
        data_whole_df = pd.DataFrame(data_whole.X, columns = data_whole.var.index, index  = data_whole.obs.index)
        data_whole_df["sample_id"] = data_whole.obs["sample_id"]
        data_whole_df["patient_id"] = data_whole.obs["patient_id"]
        data_whole_df["condition"] = data_whole.obs["condition"]
        data_whole_df['Object'] = list((self._object_column).astype('int'))
        data_whole_df['class'] = data_whole.obs["class"]
        ind_var_to_sample_id = {}
        for i,ii in zip(data_whole_df['sample_id'].astype('str'), data_whole_df["condition"]):
            ind_var_to_sample_id[i] = ii 
        data_whole_means = data_whole_df.groupby(["class","sample_id"], observed = True).mean(numeric_only = True).reset_index()
        data_whole_means['condition'] = data_whole_means['sample_id'].astype('str').replace(ind_var_to_sample_id)
        data_whole_means = pd.melt(data_whole_means, id_vars = ['condition',"class","sample_id"])
        data_whole_means_subset = data_whole_means[data_whole_means['class'] == unique_class]
        griddy = sns.FacetGrid(data_whole_means_subset,
                                palette = palette, col = 'antigen', 
                                col_wrap = col_wrap, sharey = False, sharex = False, 
                                height = 1.5, aspect = 1.75, legend_out = True)
        griddy.map_dataframe(plot_type_fx, y = "value", 
                                hue = "condition", palette = 'tab20')
        griddy.add_legend()
        griddy.tick_params("x", labelrotation = 90)
        griddy.refline(y = 0)
        griddy.figure.set_size_inches(width, height)
        griddy.figure.suptitle(f"{unique_class}", y = 1.01, size = 12)
        sns.move_legend(griddy, loc = "center right", bbox_to_anchor = legend_loc)
        for i in griddy.axes:
            i.set_title(i.get_title(), y = 1.03, size = text_size)
            i.set_ylabel(i.get_ylabel(), size = text_size)
            i.set_xlabel(i.get_xlabel(), size = text_size)
            i.set_yticks(i.get_yticks(), labels = i.get_yticklabels(), size = text_size)
            i.set_xticks(i.get_xticks(), labels = i.get_xticklabels(), size = text_size)
        self.grid_object = griddy
        if filename is not None:
            griddy.savefig(self.save_dir + f"/{filename}.png", bbox_inches = "tight") 
        plt.close()
        return griddy
    
    def whole_marker_exprs_ANOVA(self, 
                                 marker_class: str = 'All', 
                                 groupby_column: str = 'class', 
                                 variable: str = 'condition', 
                                 statistic: str = 'mean',
                                 area: bool = True,
                                 ) -> pd.DataFrame:
        '''
        Calculates statistics comparing the conditions in the experiment using ANOVA on the expression of [marker_class] markers 
        and %area of each class

        Args:
            marker_class (string): which markers / antigens to test by ANOVA. one of -- "All", "type","state", "none". 

            groupby_column (string): which column the data will be grouped by for the purposed of running separate ANOVAs for each group 
                    (as this is whole-class analysis, should always be "class")

            variable (string): which column in self.data.obs will be trated as the column containing condition / group information

            statistic (string): one of --"ANOVA", "Kruskal" -- which statistical test (ANOVA, kruskal-wallis), and what aggregate statistic 
                    (mean/std or median/IQR, respectively) is calculated & displayed in the final dataframe

            area (bool): whether to also calculate an ANOVA comparing the %area of each class between the conditions (default is True)

        Returns:
            (pandas dataframe): the pandas dataframe contianing the statistical outputs of this test.
        '''
        statistic_to_test = {"ANOVA": scipy.stats.f_oneway, "Kruskal": scipy.stats.kruskal}
        test_func = statistic_to_test[statistic]
        statistic_to_agg_func = {"ANOVA":"mean", "Kruskal":"median"}
        statistic = statistic_to_agg_func[statistic]

        data_whole = self.data.copy()
        panel = self._panel.copy()
        if marker_class != "All":
            non_type_antigens = panel[panel["marker_class"] == marker_class]['antigen']
            data_whole = data_whole[:, panel["marker_class"] == marker_class]
        else:
            non_type_antigens = panel['antigen']
    
        data_whole_df = pd.DataFrame(data_whole.X, columns = data_whole.var.index, index  = data_whole.obs.index)
        data_whole_df["sample_id"] = data_whole.obs["sample_id"]
        data_whole_df["patient_id"] = data_whole.obs["patient_id"]
        data_whole_df["condition"] = data_whole.obs["condition"]
        data_whole_df['Object'] = list((self._object_column).astype('int'))
        data_whole_df['class'] = data_whole.obs["class"]
        ind_var_to_sample_id = {}
        for i,ii in zip(data_whole_df['sample_id'].astype('str'), data_whole_df["condition"]):
            ind_var_to_sample_id[i] = ii 
        data_whole_means = data_whole_df.groupby(["class","sample_id"], observed = True).mean(numeric_only = True).reset_index()
        data_whole_means['condition'] = data_whole_means['sample_id'].astype('str').replace(ind_var_to_sample_id)
        data_whole_means = pd.melt(data_whole_means, id_vars = ['condition',"class","sample_id"])
        grand_csv = pd.DataFrame()
        whole_regionprops_dir = self.directory  + "/regionprops"
        regionprops_list = [i for i in sorted(os.listdir(whole_regionprops_dir)) if i.lower().find(".csv") != -1]
        for i in regionprops_list:
            csv = pd.read_csv(whole_regionprops_dir + "/" + i)
            grand_csv = pd.concat([grand_csv,csv], axis = 0)
        grand_csv['class'] = list(data_whole_df['class'])
        grand_csv['sample_id'] = list(data_whole_df['sample_id'])
        grand_csv['condition'] = list(data_whole_df['condition'])
        grand_csv['area'] = grand_csv['area'] / grand_csv['image_area']      ## Percentile of total image area
        ind_var_to_sample_id = {}
        for i,ii in zip(grand_csv['sample_id'], grand_csv["condition"]):
            ind_var_to_sample_id[i] = ii 
        percent_areas = grand_csv[['area','class','sample_id']].groupby(['class','sample_id'], observed = False).mean() * 100
        percent_areas = percent_areas.reset_index()
        percent_areas['sample_class'] = [i + ii  for i, ii in zip(percent_areas['class'].astype('str'),percent_areas['sample_id'].astype('str'))]
        percent_dict = {}
        for i,ii in zip(percent_areas['area'],percent_areas['sample_class']):
            percent_dict[ii] = float(i)
        percent_areas['condition'] = percent_areas['sample_id'].replace(ind_var_to_sample_id)
        data_whole_df['sample_class'] = [i + ii  for i, ii in zip(data_whole_df['class'].astype('str'),data_whole_df['sample_id'].astype('str'))]
        data_whole_df['areas******'] = data_whole_df['sample_class'].astype('str').replace(percent_dict).astype('float')
        data_whole_df = data_whole_df.drop('sample_class', axis = 1)
        data = data_whole_df.drop('Object', axis = 1)
        conditions = data[variable].unique()
        merging_clusters = data[groupby_column].unique()
        if area is True:
            non_type_antigens = list(non_type_antigens) + ['areas******']
            non_type_antigens = pd.Series(non_type_antigens, name = 'antigen')
        data_df = data.copy()
        ind_var_to_sample_id = {}
        for i,ii in zip(data_df['sample_id'].astype('str'), data_df[variable]):
            ind_var_to_sample_id[i] = ii 
        
        if statistic == "median":
            stat_column_name = "H statistic"
            data_df = data_df.groupby(['sample_id', groupby_column], observed = False).median(numeric_only = True).fillna(0).reset_index()
        elif statistic == "mean":
            stat_column_name = "F statistic"
            data_df = data_df.groupby(['sample_id', groupby_column], observed = False).mean(numeric_only = True).fillna(0).reset_index()
        
        data_df[variable] = data_df['sample_id'].astype('str').replace(ind_var_to_sample_id)
        grand_condition_list = []
        for j,jj in enumerate(merging_clusters):
            merging_data = data_df[data_df[groupby_column] == jj]
        
            condition_list = []
            for i in conditions:
                condition_data = merging_data[merging_data[variable] == i]
                condition_list.append(condition_data.drop(['sample_id',groupby_column,variable], axis = 1))
            grand_condition_list.append(condition_list)
            ANOVA_f, ANOVA_p = test_func(*condition_list)
            if j == 0:
                merging_array_F = ANOVA_f[np.newaxis,:]
                merging_array_p = ANOVA_p[np.newaxis,:]
            else:
                merging_array_F = np.concatenate([merging_array_F,ANOVA_f[np.newaxis,:]], axis = 0)
                merging_array_p = np.concatenate([merging_array_p,ANOVA_p[np.newaxis,:]], axis = 0)

        p_adj_df = pd.DataFrame(scipy.stats.false_discovery_control(np.nan_to_num(merging_array_p) + 1e-25, 
                                                                    method = 'bh'), 
                                columns = non_type_antigens)
        
        p_adj_df[groupby_column] = merging_clusters
        p_adj_df = pd.melt(p_adj_df, id_vars = groupby_column)
        p_value_df = pd.DataFrame(merging_array_p, columns = non_type_antigens)
        p_value_df[groupby_column] = merging_clusters
        p_value_df = pd.melt(p_value_df, id_vars = groupby_column)
        F_values_df = pd.DataFrame(merging_array_F, columns = non_type_antigens)
        F_values_df[groupby_column] = merging_clusters
        F_values_df = pd.melt(F_values_df, id_vars = groupby_column)
        
        final_df = pd.DataFrame()
        final_df['antigen'] = p_adj_df['antigen']
        final_df[groupby_column] = p_adj_df[groupby_column]
        final_df['p_adj'] = [sigfig.round(i, 3, warn = False) for i in p_adj_df['value']]
        final_df['p_value'] = [sigfig.round(i, 3, warn = False) for i in p_value_df['value']]     
        final_df[stat_column_name] = [sigfig.round(i, 3, warn = False) for i in F_values_df['value']]
        final_df = final_df.groupby(['antigen',groupby_column], observed = False).mean()
        condition_dict = {}
        for jj,j in enumerate(grand_condition_list):
            for i,ii in enumerate(j):
                ii[groupby_column] = data_df[groupby_column].astype('str')
                if statistic == "mean":
                    merging_mean_condition = pd.melt(ii.groupby(groupby_column, 
                                                                observed = True).mean(numeric_only = True).T.reset_index(), 
                                                     id_vars = 'antigen')
                    merging_std_condition = pd.melt(ii.groupby(groupby_column, 
                                                               observed = True).std(numeric_only = True).T.reset_index(), 
                                                    id_vars = 'antigen')
                    spread_stat = 'stdev'
                elif statistic == "median":
                    merging_mean_condition = pd.melt(ii.groupby(groupby_column, 
                                                                observed = True).median(numeric_only = True).T.reset_index(),
                                                    id_vars = 'antigen')
                    merging_std_75 = pd.melt(ii.groupby(groupby_column, 
                                                               observed = True).quantile(q = 0.75, 
                                                                                         numeric_only = True).T.reset_index(),
                                                    id_vars = 'antigen')
                    merging_std_25 = pd.melt(ii.groupby(groupby_column,
                                                        observed = True).quantile(q = 0.25,
                                                                                    numeric_only = True).T.reset_index(),
                                            id_vars = 'antigen')
                    merging_std_condition = merging_mean_condition.copy()
                    merging_std_condition['value'] = merging_std_75['value'] - merging_std_25['value']
                    spread_stat = 'IQR'
                merging_mean_condition = merging_mean_condition.groupby(['antigen',groupby_column], observed = False).mean()   
                                ## These secondary groupby lines only serve to set up the multi-index (values should not change)
                merging_std_condition = merging_std_condition.groupby(['antigen',groupby_column], observed = False).mean()
                merging_mean_condition['stdev'] = merging_std_condition['value']
                if jj == 0:
                    condition_dict[str(i)] = merging_mean_condition
                else:
                    condition_dict[str(i)] = pd.concat([condition_dict[str(i)], merging_mean_condition], axis = 0)
                
        for i,ii in enumerate(condition_dict):
            final_df[f'avg {conditions[i]} {statistic} exprs'] = condition_dict[ii]['value']   
            final_df[f'avg {conditions[i]} {statistic} exprs'] = [sigfig.round(i, 3, warn = False) for i in final_df[f'avg {conditions[i]} {statistic} exprs']]
            final_df[f'{conditions[i]} {spread_stat}'] = condition_dict[ii]['stdev']
            final_df[f'{conditions[i]} {spread_stat}']  = [sigfig.round(i, 3, warn = False) for i in final_df[f'{conditions[i]} {spread_stat}']]
        self.cluster_expression_ANOVA_df = final_df.reset_index().sort_values(stat_column_name, ascending = False)
        return final_df.reset_index().sort_values(stat_column_name, ascending = False)

    def plot_heatmap(self, 
                     type_of_stat: str = "F statistic",
                     filename: Union[str, None] = None,
                     ) -> plt.figure:
        '''
        Plots a statistics heatmap. -Neg log(statistic) if the statistic is a p value instead of an F statistic
        '''
        heatmap_df = self.cluster_expression_ANOVA_df.pivot(index = 'antigen', columns = 'class', values = type_of_stat)
        heatmap_df = heatmap_df.astype('float')
        title_helper = ''
        if not ((type_of_stat == "F statistic") or (type_of_stat == "H statistic")):
            heatmap_df = heatmap_df.transform(np.log) * (-1)
            title_helper = "Neg log "
        figure2 = plt.figure()
        ax = plt.gca()
        sns.heatmap(heatmap_df, cmap = "coolwarm", ax = ax, linewidths = 0.01, vmin = 0, vmax = 7, xticklabels = True, yticklabels = True)
        ax.set_title(f"{title_helper}{type_of_stat} Heatmap")
        if filename is not None:
            figure2.savefig(self.save_dir + f"/{filename}.png", bbox_inches = "tight") 
        plt.close()
        return figure2
    
    #### shared with the main experiment class
    def export_data(self, 
                    subset_columns: Union[list[str], None] = None, 
                    subset_types: Union[list[list[str]], None] = None, 
                    groupby_columns: Union[list[str], None] = None, 
                    statistic: str = 'mean',
                    groupby_nan_handling: str = 'zero',
                    include_marker_class_row: bool = False,
                    untransformed: bool = False,
                    filename: Union[str, None] = None, 
                    ) -> pd.DataFrame:
        '''
        Exports currently loaded data from the Analysis, from self.data. 
        
        Preserves any previously performed scaling, dropped categories, & batch correction. Always of arcsinh(data / 5) transformed data. Can
        export the entirety of relevant self.data information, or export subsets of self.data, and/or export aggregate summary statistics for 
        groups within the data. 

        Args:
            subset_columns (list[str] or None): 
                a list of strings denoting the columns to subset the dataframe's rows on (here and in other arguments, non-string input is attepmted 
                to be cast to strings inside the function, as well as the corresponding column of the data). if this or subset_types is None, no subsetting occurs. 

            subset_types (list[list[str]] or None): 
                a list contains sub-lists for strings. The length of the upper list must be the length of
                the subset_columns list, as each sub-list contains strings corresponding to the rows to keep. 

                    As in: if subset_columns = ['column1', 'column3'] and subset_types = [['type2', 'type6'],['typeB', 'typeZ']],
                    then rows of type2 / type6 in column1 will be kept, and similarly rows of typeB / typeZ in column2.

                When > 1 columns / conditions are subsetted on, as in the above example, the rows that are kept are the union of 
                all the subsetting conditions WITHIN a given column, but the intersection BETWEEN what is kept from each column. 
                So in the above example, all rows of column1 == type2/6 that also have column2 == typeB/Z are the rows that are maintained.
                                                        
            groupby_columns (list[str] or None): 
                A list of strings indicating what columns of the data to groupby. If None, then grouping is not performed.
                Used like this:    self.data.obs.groupby(groupby_columns)              but on a dataframe containing the data.X values as well

            statistic (str): 
                Possible values: 'mean','median','sum','std','count'. Denotes the pandas groupby method to be used after grouping (ignored if groupby_columns is None).
                Numeric methods (mean, median, sum, std) are only applied to numeric columns, so only those columns + the groupby columns 
                will be in the final dataframe / csv
            
            groupby_nan_handling(str):
                'zero' or 'drop' -- when grouping the data whether to drop (nans), which usually represent non-existent category combinations or to 
                convert nans to zeros. Any other values of this parameter will cause NaNs to be left as-is in the data export
                Note that the default (and only option available in GUI) is 'zero', which converts ALL NaN values to 0, while the 'drop' option only drops
                rows where EVERY numerical value is NaN.
                By default, all possible groupby_columns combinations are included in the export (even if they are not present in the data, such cell types 
                not present in every ROI), This is the source of most NaN values. Notably, columnns in the metadata (not data.obs!) of the Analysis are given special 
                treatment to try to prevent non-existent experimental categories from having data exported (for example, each ROI / sample_id should have been 
                with a single condition, not every possible condition in the dataset). 

            include_marker_class_row (bool): 
                Whether to include the marker_class information as a row at the bottom of the table --> True to 
                include this row -- useful for reimport into PalmettoBUG.
                False to not include this row -- this is probably better for import into non-PalmettoBUG software for analysis,
                or at the least the user will need to remember to remove this row before analyzing!
                When the marker_class row is included, it is encoded as integers (to prevent mixed dtype issues/warnings on reload)
                
                    >>> 0 = 'none', 1 = 'type', 2 = 'state'

                metadata columns (which have no marker_class) have this row filled with 'na'. 
                NOT USED IN COMBINATION WITH GROUPING!

            untransformed (bool):
                if True, will export the untransformed (pre-arcsinh, pre-scaling, etc., etc.) data, from self.data.uns['count'].
                Provided so that the raw data is not difficult to recover, although not expected to be used frequently. Default == False. 

            filename: (str, or None): 
                the name of the csv file to save the exported dataframe inside the self.data_table_dir folder. If None, no export occurs, and the data table is only returned. 

        Returns:
            (pandas DataFrame) -- the pandas dataframe representing the exported data. 

        Inputs/Outputs:
            Outputs: 
                If filename is provided (is not None), then exports the data table to self.data_table_dir/filename.csv
        '''
        if filename is None:
            output_path =  None
        else:
            output_path = "".join([self.data_table_dir, "/", str(filename), ".csv"])
        data = self.data.copy()
        if untransformed:
            data.X = self.data.uns['counts'].copy()
        data.obs = data.obs.reset_index()    ## cell index included as 'index' column --> useful if dropping / filtering cells out of the dataset
        ## anndata to pd.DataFrame:
        data_points = pd.DataFrame(data.X)
        data_points.columns = data.var.index.astype('str')
        data.obs.columns = data.obs.columns.astype('str')
        data_col_list = [i for i in data.obs.columns]
        for i in data_col_list:
            if (include_marker_class_row is True) & (groupby_columns is None):
                data_points[str(i)] = list(data.obs[str(i)]) + ["na"]
            else:
                data_points[str(i)] = list(data.obs[str(i)])
                data_points[str(i)] = data_points[str(i)].astype(data.obs[str(i)].dtype)
        
        # subset:
        if (subset_types is not None) and (subset_columns is not None):
            counter1 = 0
            for i,ii in zip(subset_columns, subset_types):
                counter2 = 0
                for j in ii:
                    data_slice = data_points[data_points[str(i)].astype('str') == str(j)]    ## subsets for
                    if counter2 == 0:
                        new_data = data_slice.copy()
                        counter2 = 1
                    else:
                        new_data = new_data.merge(data_slice, how = 'outer')
                if counter1 == 0 :
                    counter1 = 1
                    output_df = new_data.copy()
                else:
                    output_df = output_df.merge(new_data, how = 'inner')
            data_points = output_df
            
        if groupby_columns is None:
            if output_path is not None:
                data_points.to_csv(str(output_path), index = False)
            return data_points

        else:
            extra_columns = None
            if len(groupby_columns) > 1:
                extra_columns = [i for i in groupby_columns if i in self.metadata.columns]
                if len(extra_columns) > 1:
                    extra_data_points = data_points[extra_columns]
                    def concat(*args):
                        return "_|_|_".join(*args)
                    data_points['use'] = extra_data_points.T.apply(concat)
                    groupby_columns = ['use'] + [i for i in groupby_columns if i not in self.metadata.columns]
                else:
                    extra_columns = None
      
            groupby_object = data_points.groupby(groupby_columns, observed = False)
    
            if statistic == 'mean':
                groupby_object = groupby_object.mean(numeric_only = True)
            if statistic =='median':
                groupby_object = groupby_object.median(numeric_only = True)
            if statistic =='sum':
                groupby_object = groupby_object.sum(numeric_only = True)
            if statistic =='std':
                groupby_object = groupby_object.std(numeric_only = True)
            if statistic =='count':
                groupby_object = groupby_object.count()
                groupby_object = pd.DataFrame(groupby_object[groupby_object.columns[0]])
                groupby_object.columns = ['count']
    
            groupby_object = groupby_object.reset_index()
            if statistic == 'count':
                pass
            else:
                backup_groupby = pd.DataFrame(groupby_object[groupby_columns], index = groupby_object.index)
                if groupby_nan_handling == 'drop':
                    groupby_object = groupby_object.drop(groupby_columns, axis = 1).dropna(how = 'all')
                elif groupby_nan_handling == 'zero':
                    groupby_object = groupby_object.drop(groupby_columns, axis = 1).fillna(0)

                groupby_object = pd.concat([backup_groupby, groupby_object], axis = 1)

            for i in data_col_list:
                if (i in groupby_object.columns.astype('str')) and (i not in groupby_columns):
                    groupby_object = groupby_object.drop(i, axis = 1)
            if extra_columns:
                def split(value, part = 0):
                    return value.split("_|_|_")[part]
                interim = pd.DataFrame()
                for i,ii in enumerate(extra_columns):
                    interim[ii] = groupby_object['use'].apply(split, part = i)
                groupby_object = pd.concat([interim, groupby_object], axis = 1)
                groupby_object = groupby_object.drop('use', axis = 1)
    
    
            if output_path is not None:
                groupby_object.to_csv(str(output_path), index = False)
    
            return groupby_object


'''
This module contains any special exceptions to be used in the PalmettoBUG program

Like the rest of PalmettoBUG, this code is licensed under the GPL-3 open source license.

'''

__all__ = []

class NoSharedFilesError(Exception):
    '''
    This Error occurs when searching through two parallel directories, which are expected to have matching filenames in each directory,
    but the two directories do not shared any common filenames
    '''


'''
"vendorizing" the following package: https://github.com/Maarten-vd-Sande/qnorm

(
MIT License

Copyright (c) 2020, Maarten van der Sande

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
)


Changes:
-- Only focused on the functions that I use in PalmettoBUG (quantile_normalize) -- remove anything not needed for that
-- Joined all files into one (.util import in the original ---> .util functions moved into this file)
-- Removed multiple dispatch (I only pass in numpy arrays to qnorm) and pandas check
-- Removed unused imports
-- add __all__ for docs

'''
__all__ = []
from typing import Union
from multiprocessing import Pool, RawArray

import numba
import numpy as np


def _parallel_argsort(_array, ncpus, dtype):
    """
    private argsort function of qnorm that works with multiple cpus
    """
    # multiproces sorting
    # first we make a shared array
    data_shared = RawArray(
        np.ctypeslib.as_ctypes_type(dtype), _array.shape[0] * _array.shape[1]
    )
    # and wrap it with a numpy array and fill it with our data
    data = np.frombuffer(data_shared, dtype=dtype).reshape(_array.shape)
    np.copyto(data, _array.astype(dtype))

    # now multiprocess sort
    with Pool(
        processes=ncpus,
        initializer=_worker_init,
        initargs=(data_shared, dtype, data.shape),
    ) as pool:
        sorted_idx = np.array(
            pool.map(_worker_sort, range(data.shape[1])), dtype=np.int64
        ).T
    return data, sorted_idx

var_dict = {}
def _worker_init(X, X_dtype, X_shape):
    """
    helper function to pass our reference of X to the sorter
    """
    var_dict["X"] = X
    var_dict["X_dtype"] = X_dtype
    var_dict["X_shape"] = X_shape


def _worker_sort(i):
    """
    argsort a single axis
    """
    X_np = np.frombuffer(var_dict["X"], dtype=var_dict["X_dtype"]).reshape(
        var_dict["X_shape"]
    )
    return np.argsort(X_np[:, i])

def quantile_normalize(
    _data: np.ndarray,
    axis: int = 1,
    target: Union[None, np.ndarray] = None,
    ncpus: int = 1,
) -> np.ndarray:
    # check for supported dtypes
    if not np.issubdtype(_data.dtype, np.number):
        raise ValueError(
            f"The type of your data ({_data.dtype}) is is not "
            f"supported, and might lead to undefined behaviour. "
            f"Please use numeric data only."
        )
    # numba does not (yet) support smaller
    elif any(
        np.issubdtype(_data.dtype, dtype) for dtype in [np.int32, np.float32]
    ):
        dtype = np.float32
    else:
        dtype = np.float64

    # take a transposed view of our data if axis is one
    if axis == 0:
        _data = np.transpose(_data)
    elif axis == 1:
        pass
    else:
        raise ValueError(
            f"qnorm only supports 2 dimensional data, so the axis"
            f"has to be either 0 or 1, but you set axis to "
            f"{axis}."
        )

    # sort the array, single process or multiprocessing
    if ncpus == 1:
        # single process sorting
        data = _data.astype(dtype=dtype)
        # we do the sorting outside of numba because the numpy implementation
        # is faster, and numba does not support the axis argument.
        sorted_idx = np.argsort(data, axis=0)
    elif ncpus > 1:
        # multiproces sorting
        data, sorted_idx = _parallel_argsort(_data, ncpus, dtype)
    else:
        raise ValueError("The number of cpus needs to be a positive integer.")

    sorted_val = np.take_along_axis(data, sorted_idx, axis=0)

    if target is None:
        # if no target supplied get the (sorted) rowmeans
        target = np.mean(sorted_val, axis=1)
    else:
        # otherwise make sure target is correct data type and shape
        if not isinstance(target, np.ndarray):
            try:
                target = np.array(target)
            except Exception:
                raise ValueError(
                    "The target could not be converted to a " "numpy.ndarray."
                )
        if target.ndim != 1:
            raise ValueError(
                f"The target array should be a 1-dimensionsal vector, however "
                f"you supplied a vector with {target.ndim} dimensions"
            )
        if target.shape[0] != data.shape[0]:
            raise ValueError(
                f"The target array does not contain the same amount of values "
                f"({target.shape[0]}) as the data contains rows "
                f"({data.shape[0]})"
            )
        if not np.issubdtype(target.dtype, np.number):
            raise ValueError(
                f"The type of your target ({data.dtype}) is is not "
                f"supported, and might lead to undefined behaviour. "
                f"Please use numeric data only."
            )
        target = np.sort(target.astype(dtype=dtype))

    final_res = _numba_accel_qnorm(data, sorted_idx, sorted_val, target)
    if axis == 0:
        final_res = final_res.T
    return final_res


@numba.jit(nopython=True, fastmath=True, cache=True)
def _numba_accel_qnorm(
    qnorm: np.ndarray,
    sorted_idx: np.ndarray,
    sorted_val: np.ndarray,
    target: np.ndarray,
) -> np.ndarray:
    """
    numba accelerated "actual" qnorm normalization.
    """
    # get the shape of the input
    n_rows = qnorm.shape[0]
    n_cols = qnorm.shape[1]

    for col_i in range(n_cols):
        i = 0
        # we fill out a column not from lowest index to highest index,
        # but we fill out qnorm from lowest value to highest value
        while i < n_rows:
            n = 0
            val = 0.0
            # since there might be duplicate numbers in a column, we search for
            # all the indices that have these duplcate numbers. Then we take
            # the mean of their rowmeans.
            while (
                i + n < n_rows
                and sorted_val[i, col_i] == sorted_val[i + n, col_i]
            ):
                val += target[i + n]
                n += 1

            # fill out qnorm with our new value
            if n > 0:
                val /= n
                for j in range(n):
                    idx = sorted_idx[i + j, col_i]
                    qnorm[idx, col_i] = val

            i += n

    return qnorm